"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
from pyEELSMODEL.core.component import Component
from pyEELSMODEL.core.parameter import Parameter
import numpy as np
import logging

logger = logging.getLogger(__name__)


class Polynomial(Component):
    """
    A Polynomial component. The constant value of the polynomials
    are all set to 1 as initial value. The coefficients of the powers are
    normalized since they produce numerical results because the x-axis
    is useually expressed in eV.

    Parameters
    ----------
    specshape : Spectrumshape
        The spectrum shape used to model
    order : uint
        The order of the polynomial.

    Returns
    -------
    """

    def __init__(self, specshape, order=1):
        """

        :param specshape:
        :param order:
        """
        super().__init__(specshape)
        self.setname('Polynomial')
        self.order = order + 1

        for i in range(self.order):
            name = 'x' + str(self.order - i - 1)
            # changed
            p = Parameter(name, 1)
            p.setlinear(True)
            p.sethasgradient(True)
            self._addparameter(p)

        self._setname('Polynomial')

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        is_changed = False
        for param in self.parameters:
            if param.ischanged():
                is_changed = True

        if is_changed:
            coeff = []
            for param in self.parameters:
                coeff.append(param.getvalue())

            # normalize the coefficients since they will produce
            # numerical problems
            coeffn = np.zeros(len(coeff))
            for i in range(coeffn.size):
                coeffn[i] = coeff[i] / (
                            self.energy_axis ** (self.order - i - 1)).sum()

            p = np.poly1d(coeffn)
            self.data = p(self.energy_axis)
        self.setunchanged()  # put parameters to unchanged

    def getgradient(self, parameter):
        for index, param in enumerate(self.parameters):
            if param is parameter:
                param_order = len(self.parameters) - (index + 1)

        if param_order is None:
            logger.info('Parameter is non existing in this component')

        return self.energy_axis ** (param_order)



from pyEELSMODEL.core.operator import Operator
from pyEELSMODEL.core.model import Model
from pyEELSMODEL.core.multispectrum import MultiSpectrum, MultiSpectrumshape
from pyEELSMODEL.core.spectrum import Spectrum
from pyEELSMODEL.components.gaussian import Gaussian
from pyEELSMODEL.components.lorentzian import Lorentzian
from pyEELSMODEL.components.voigt import Voigt
from pyEELSMODEL.fitters.lsqfitter import LSQFitter
from pyEELSMODEL.operators.multispectrumvisualizer import \
    MultiSpectrumVisualizer
import logging
import numpy as np
import matplotlib.pyplot as plt
logger = logging.getLogger(__name__)


class ZLPRemoval(Operator):
    """
    Removes the ZLP from the spectrum by fitting the appropriate function
    to it
    """

    def __init__(self, spectrum, signal_range=None, model_type='Gaussian'):
        self.spectrum = spectrum
        self.signal_range = signal_range
        self.model_type = model_type
        self.fwhm_boundaries = (self.spectrum.dispersion/2,
                                40*self.spectrum.dispersion)

        self.set_indices()
        self.make_zeroloss_model()
        self.set_start_parameters()

        self.start_parameters = None
        self.fitter = None
        self.model_signal = None

        self.inelastic = None
        self.zlp = None
        self.mirror_zlp = None
        self.mirror_inelastic = None

        self.mirror_ind = 50

    @property
    def spectrum(self):
        return self._spectrum

    @spectrum.setter
    def spectrum(self, spectrum):
        self._spectrum = spectrum

    @property
    def signal_range(self):
        return self._signal_range

    @signal_range.setter
    def signal_range(self, signal_range):
        if signal_range is None:
            # todo a smarter guess of the signal range would be good
            signal_range = (self.spectrum.offset,
                            self.spectrum.energy_axis[-1])
        self._signal_range = signal_range

    @property
    def model_type(self):
        return self._model_type

    @model_type.setter
    def model_type(self, model_type):
        self._model_type = model_type

    @property
    def model(self):
        """
        The model used for the fitting
        """
        return self._model

    @model.setter
    def model(self, m0):
        """
        Sets the attribute model to the given model.
        """
        self._model = m0

    @property
    def fitter(self):
        """
        The fitter used in the spectrum (now it is the LSQ since
        it fast and stable)
        """
        return self._fitter

    @fitter.setter
    def fitter(self, fit):
        self._fitter = fit

    def make_zeroloss_model(self):
        """
        Creates a model for the zero loss peak, this depends on which
        model_type is chosen when creating the background object. The model is
        stored in the model attribute
        """
        specshape = self.spectrum.get_spectrumshape()
        m0 = Model(specshape)
        if self.model_type == 'Gaussian':
            comp = Gaussian(specshape, A=1, centre=0, fwhm=1)
        elif self.model_type == 'Lorentzian':
            comp = Lorentzian(specshape, A=1, centre=0, fwhm=1)
        elif self.model_type == 'Voigt':
            comp = Voigt(specshape, A=1, centre=1, gamma=1, sigma=1)
        else:
            print('model type not included in the list of possibilities')

        comp.parameters[1].setboundaries(self.signal_range[0],
                                         self.signal_range[1], force=True)
        comp.parameters[2].setboundaries(self.fwhm_boundaries[0],
                                         self.fwhm_boundaries[1])
        m0.addcomponent(comp)

        self.model = m0

    def estimate_start_param(self):
        start_param = np.zeros(self.model.getnumparameters())
        ind0 = self.indices[0]
        ind1 = self.indices[1]

        start_param[0] = np.max(self.spectrum.data[ind0:ind1])
        co_max = np.argmax(self.spectrum.data[ind0:ind1])+ind0
        start_param[1] = self.spectrum.energy_axis[int(co_max)]
        start_param[2:] = self.spectrum.dispersion*10
        self.start_param = start_param

    def estimate_start_param_multi(self):
        """
        Estimates the starting value of the fitting for a multispectrum when
        having a gaussian or lorentzian model to fit.
        """
        # The lorentzian and gaussian model both have three parameters.
        start_param = np.zeros((self.spectrum.xsize, self.spectrum.ysize,
                                self.model.getnumparameters()))
        ind0 = self.indices[0]
        ind1 = self.indices[1]

        start_param[:, :, 0] = np.max(self.spectrum.multidata, axis=2)

        # this should be zero in principle
        co = np.argmax(self.spectrum.multidata[:, :, ind0:ind1], axis=2)
        shift = (co + ind0)
        start_param[:, :, 1] = self.spectrum.energy_axis[shift]

        # if voigt this is still  fine
        # start_param[:, :, 2:] = 4*self.spectrum.dispersion
        start_param[:, :, 2:] = 1

        self.start_param_multi = start_param

    def set_start_parameters(self):
        if isinstance(self.spectrum, MultiSpectrum):
            self.estimate_start_param_multi()
        else:
            self.estimate_start_param()

    def set_indices(self):
        """
        Calculates the indices used which are excluded in the fit.
        These indices are also used to determine a first guess of the
        background model. The result is stored in the indices attribute

        """
        ind1 = [self.spectrum.get_energy_index(self.signal_range[0]),
                self.spectrum.get_energy_index(self.signal_range[1])]
        self.indices = ind1

    def include_areas(self):
        """
        Sets the exlude of the spectrum such that only the integration
        range is taken into account.
        """
        self.spectrum.set_include_region(self.indices[0], self.indices[1])

    def fit_zlp(self):
        prev_exclude = self.spectrum.exclude[:]
        self.spectrum.exclude = np.ones(self.spectrum.size, dtype=bool)
        self.include_areas()

        fit = LSQFitter(self.spectrum, self.model, use_bounds=True)
        if isinstance(self.spectrum, MultiSpectrum):
            fit.multi_fit(start_param=self.start_param_multi)
            self.fitter = fit
            self.zlp = fit.model_to_multispectrum()
        else:
            for ii, param in enumerate(self.model.getfreeparameters()):
                print(self.start_param[ii])
                param.setvalue(self.start_param[ii])

            fit.perform_fit()
            self.fitter = fit
            self.zlp = Spectrum(self.spectrum.get_spectrumshape(),
                                data=self.model.data)

        self.spectrum.exclude = prev_exclude
        self.inelastic = self.spectrum - self.zlp

    def mirrored_zlp(self):
        ind0 = self.indices[0]
        ind1 = self.indices[1]
        print(isinstance(self.spectrum, MultiSpectrum))
        if isinstance(self.spectrum, MultiSpectrum):
            shift = (np.argmax(self.spectrum.multidata[:, :, ind0:ind1],
                               axis=2)+ind0)
            shape = (self.spectrum.xsize, self.spectrum.ysize)
            ndata = np.zeros((self.spectrum.xsize, self.spectrum.ysize,
                              self.spectrum.size))
            for index in (np.ndindex(shape)):
                isl = np.s_[index]
                ind = max(0, shift[isl]-self.mirror_ind)
                res = self.spectrum.multidata[isl[0], isl[1], ind:shift[isl]+1]

                rest_size = int(self.spectrum.size - 2*res.size - ind + 1)
                ndata[isl] = np.concatenate((np.zeros(ind),
                                             res, np.flip(res[:-1]),
                                             np.zeros(rest_size)))

            mshape = MultiSpectrumshape(self.spectrum.dispersion,
                                        self.spectrum.offset,
                                        self.spectrum.size,
                                        self.spectrum.xsize,
                                        self.spectrum.ysize)
            self.mirror_zlp = MultiSpectrum(mshape, data=ndata)
            self.mirror_inelastic = self.spectrum - self.mirror_zlp

        else:
            shift = np.argmax(self.spectrum.data[ind0:ind1])+ind0
            ind = max(0, shift - self.mirror_ind)
            print(ind)
            res = self.spectrum.data[ind:shift+1]
            rest_size = int(self.spectrum.size - 2*res.size - ind + 1)
            ndata = np.concatenate((np.zeros(ind),
                                    res, np.flip(res[:-1]),
                                    np.zeros(rest_size)))
            self.mirror_zlp = Spectrum(self.spectrum.get_spectrumshape(),
                                       data=ndata)
            self.mirror_inelastic = self.spectrum - self.mirror_zlp

    def show_fit_result(self):
        if isinstance(self.spectrum, MultiSpectrum):
            print('show the multispectrum')
            MultiSpectrumVisualizer([self.spectrum, self.zlp],
                                    labels=['Experimental', 'Fitted ZLP'])
        else:
            self.fitter.plot()

    def show_mirror_result(self):
        if isinstance(self.spectrum, MultiSpectrum):
            print('show the multispectrum')
            MultiSpectrumVisualizer([self.spectrum, self.zlp],
                                    labels=['Experimental', 'Fitted ZLP'])
        else:
            fig, ax = plt.subplots()
            ax.plot(self.spectrum.energy_axis, self.spectrum.data,
                    color='black', marker='o')
            ax.plot(self.spectrum.energy_axis, self.mirror_zlp.data)



from pyEELSMODEL.core.component import Component
from pyEELSMODEL.core.parameter import Parameter
from pyEELSMODEL.components.CLedge.coreloss_edge import CoreLossEdge
import numpy as np
from scipy.interpolate import interp1d
import logging

logger = logging.getLogger(__name__)


class GDOSLin(Component):
    """
    Linear fine structure DOS Component
    This component will create in the onset region [estart-estart+ewidth] of a
    core loss a interpolated curve going through degree points which are free
    parameters of the model
    This curve is then added to the core loss cross section and will increase
    or decrease certain areas in the cross section
    As this is added, this is also a linear component

    """

    def __init__(self, specshape, estart, ewidth=50, degree=20,
                 interpolationtype='linear'):
        """
        Initializes the gdoslin object.

        Parameters
        ----------
        specshape : Spectrumshape
            The spectrum shape used to model

        estart: float [eV]
            The onset value of the fine structure. This is mostly
            the onset energy of the edge on wich you want to apply the
            fine structure.

        ewidth : float [eV]
            The energy width over which the fine structure is modelled.
            (default: 50)

        degree: int
            The number of parameters used to model the fine structure.
            (default: 20)

        interpolationtype: string
            The type of interpolation used.

        Returns
        -------

        """
        super().__init__(specshape)
        self.setdescription(
            "Linear Fine Structure used to alter atomic cross sections"
            " in the near edge region")

        self.degree = degree
        self.interpolationtype = interpolationtype

        p0 = Parameter("Estart", estart, False)
        self._addparameter(p0)

        p1 = Parameter("Ewidth", ewidth, False)
        self._addparameter(p1)

        # and a list of variables controlling the shape of the fine struct
        for i in range(self.degree):
            pname = 'a' + str(i)
            p = Parameter(pname, 1.0, True)
            p.setboundaries(-np.inf, np.inf)
            p.setlinear(True)
            self._addparameter(p)

        self.ctes = np.ones(self.degree + 2)
        self.initDOS()

        self.connected_edge = None
        self.set_gdos_name()

        # normalization such that each fine structure has the same integral
        self.calculate_integral_per_parameter()

    @classmethod
    def gdoslin_from_edge(cls, specshape, component, pre_e=5, ewidth=50,
                          degree=20, interpolationtype='linear'):
        """
        Class method is made to create an gdoslin which is connected to a
        coreloss edge.
        No need to input the onset energy

        Parameters
        ----------
        specshape : Spectrumshape
            The spectrum shape used to model

        component: CoreLossEdge
            The coreloss edge on which to add the fine structure.

        pre_e: float [eV]
            The value of the energy onset of the fine structure with respect
            to the onset energy of the coreloss edge. Hence for oxygen at 532eV
            with pre_e=5, the starting energy for the fine structure is 527 eV.
            (default: 5)

        ewidth : float [eV]
            The energy width over which the fine structure is modelled.
            (default: 50)

        degree: int
            The number of parameters used to model the fine structure.
            (default: 20)

        interpolationtype: string
            The type of interpolation used.

        Returns
        -------

        """

        if not isinstance(component, CoreLossEdge):
            raise TypeError(r'Component should be a CoreLossEdge')

        estart = component.onset_energy - pre_e
        comp = GDOSLin(specshape, estart, ewidth=ewidth, degree=degree,
                       interpolationtype=interpolationtype)
        comp.connected_edge = component
        # print(comp.connected_edge)
        comp.set_gdos_name()
        return comp

    def set_gdos_name(self):
        """
        Set the proper name of the gdos
        """
        s1 = "Linear Fine Structure (DOS):"
        s2 = 'Onset' + str(self.parameters[0].getvalue()) + 'eV'
        s3 = 'Width' + str(self.parameters[1].getvalue()) + 'eV: '
        if self.connected_edge is None:
            self._setname(' '.join([s1, s2, s3]))
        else:
            self._setname(' '.join([s1, s2, s3, self.connected_edge.name[:2]]))

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        p0 = self.parameters[0]
        p1 = self.parameters[1]
        Estart = p0.getvalue()
        Ewidth = p1.getvalue()
        Estop = Estart + Ewidth

        # calculate the DOS
        changes = False

        for i in range(self.degree):
            p = self.parameters[i + 2]
            changes = changes or p.ischanged()

        if p0.ischanged() or changes or p1.ischanged():
            logger.debug("parameters changed calculating DOS. degree: %d",
                         self.degree)
            logger.debug("Estart: %e", Estart)
            logger.debug("Estop: %e", Estop)

            en = self.energy_axis

            self.initDOS()

            yp = self.yp / self.ctes
            # yp = self.yp

            f = interp1d(self.Ep, yp, kind=self.interpolationtype)
            self.data[self.indexstart:self.indexstop] = f(
                en[self.indexstart:self.indexstop])

            # set parameters as unchanged since last time we calculated
            self.setunchanged()

        else:
            logger.debug(
                "parameters have not changed, i don't need to calculate again")

    def initDOS(self):
        # prepare all for the DOS to work with the given options
        p0 = self.parameters[0]
        p1 = self.parameters[1]
        Estart = p0.getvalue()
        Ewidth = p1.getvalue()
        Estop = Estart + Ewidth
        en = self.energy_axis
        self.indexstart = self.get_energy_index(Estart)
        if self.get_energy_index(Estop) == self.size:
            self.indexstop = self.get_energy_index(Estop) - 1
        else:
            self.indexstop = self.get_energy_index(Estop)
        if self.indexstop <= self.indexstart:
            logger.warning('Estart and Estop are too close together')
            return

        self.Ep = []
        self.yp = []
        Estep = (en[self.indexstop] - en[self.indexstart]) / (self.degree + 1)
        # start with yp=0 at estart
        self.Ep.append(en[self.indexstart])
        self.yp.append(0)
        for i in range(self.degree):
            self.Ep.append(en[self.indexstart] + (i + 1) * Estep)
            self.yp.append(self.parameters[i + 2].getvalue())
        self.Ep.append(en[self.indexstop])
        self.yp.append(0)
        self.yp = np.array(self.yp)

    def couple_parameter_to_edge(self):
        """
        Couples the parameters of the gdoslin to the edge such that this
        template edge can be used to fit the spectrum where the gdoslin is no
        variable anymore.
        """
        if self.connected_edge is None:
            print('no edge is connected to this gdoslin')
            return

        for i in range(self.degree):
            p = self.parameters[i + 2]
            fraction = p.getvalue() / self.connected_edge.parameters[
                0].getvalue()
            p.couple(self.connected_edge.parameters[0], fraction=fraction)

    def calculate_integral_per_parameter(self):
        """
        Since the integral for each individual parameter is not the same we
        calculate the difference which is then applied to the component
        as a multiplication factor to resolve this interpolation problem
        """
        ctes = np.ones(len(self.parameters))

        for ii, param in enumerate(self.parameters[2:]):
            for par in self.parameters[2:]:
                if par == param:
                    par.setvalue(1)
                else:
                    par.setvalue(0)
            self.calculate()

            # ctes[ii+1] = self
            ctes[ii + 1] = self.data.sum()
            ctes[ii + 1] /= (self.Ep[ii + 2] - self.Ep[ii]) / (
                        self.Ep[2] - self.Ep[0])

        self.ctes = ctes
        # print('done')

        for param in self.parameters[2:]:
            param.setvalue(1)

        self.calculate()



import numpy as np
import matplotlib.pyplot as plt
import logging
from pyEELSMODEL.core.operator import Operator
from pyEELSMODEL.core.multispectrum import MultiSpectrum
from pyEELSMODEL.core.model import Model

from pyEELSMODEL.operators.aligns.fastalignzeroloss import FastAlignZeroLoss

from pyEELSMODEL.components.powerlaw import PowerLaw
from pyEELSMODEL.components.linear_background import LinearBG
from pyEELSMODEL.components.CLedge.zezhong_coreloss_edgecombined import\
    ZezhongCoreLossEdgeCombined
from pyEELSMODEL.components.CLedge.kohl_coreloss_edgecombined import \
    KohlLossEdgeCombined
from pyEELSMODEL.components.MScatter.mscatterfft import MscatterFFT
from pyEELSMODEL.components.gdoslin import GDOSLin

from pyEELSMODEL.fitters.linear_fitter import LinearFitter
from pyEELSMODEL.fitters.lsqfitter import LSQFitter

logger = logging.getLogger(__name__)


class ElementalQuantification(Operator):
    """
    Object which has the workflow to perform elemental quantification.
    Many different attributes can be easily modified to have some flexibility
    in the workflow.


    """
    def __init__(self, spectrum, elements, edges, settings, ll=None):
        """
        Makes an ElementalQuantification object.

        Parameters
        ----------
        spectrum: Spectrum or MultiSpectrum
            The multispectrum form which an area need to be selected
        elements: list of strings
            The elements list should contain the elements as indicated on the
            periodic table
        edges: list of strings
            The edges list should contain the letters of the edges
            (K, L, M, ...)
        settings: tuple
            Tuple containing E0 (acceleration voltage), alpha (convergence
             angle) and beta (collection angle).
        ll: Spectrum or MultiSpectrum
            The low loss which can be added. This will be used to align the
            spectra with each other and will be used to model the multiple
            scattering.

        """
        self.spectrum = spectrum
        self.is_multispectrum = isinstance(spectrum, MultiSpectrum)
        self.ll = ll

        self.do_align = True

        # show feedback
        self.feedback = True  # inidicates if inbetween results are shown

        # attributes connected to core loss components
        self.elements = elements
        self.edges = edges
        # onset is defined with respect to the energy of the edge
        self.onsets = np.zeros((len(elements)))

        self.E0 = settings[0]
        self.alpha = settings[1]
        self.beta = settings[2]
        self.qsteps = 20  # the number of points used to integrate q vector

        # attributes connected to fine structure
        self.use_fine = False  # indicates if fine structure will be used.
        self.fine_components = []
        self.fine_intervals = []
        # how much the fine structure should be fitter before onset energy
        self.pre_fine = None
        self.dE = None

        # attributes connected to background
        self.background_model = 'linear'
        self.n_bgterms = 4  # number of terms in the linear background

        # attributes connected to fitter
        self.linear_fitter_method = 'nnls'
        self.use_weights = False

        # which GOS array to use: zhang or segger
        self.gos_array = 'zhang'

        if ll is not None:
            self.estimate_sampling()

    def align_zlp(self):
        """
        Align the multispectra using the low loss. If no multispectrum is
        used, then this method will not do anything.
        """
        if self.ll is None:
            print('No low loss is added to the object so zlp alignement is '
                  'impossible')
            return None

        if not self.is_multispectrum:
            print('zlp alignment is only performed on multispectra')
            return None

        align = FastAlignZeroLoss(self.ll, other_spectra=[self.spectrum])
        align.perform_alignment()
        # align.show_shift()

        self.align = align
        self.spectrum = align.aligned_others[0]
        self.ll = align.aligned

    def _get_onset_energies(self):
        """
        Calculates the onset energies of the different edges. Is needed in
        the estimation of the autofit when using the power-law as background
        """
        onset_energies = np.zeros(len(self.element_components))
        for ii, comp in enumerate(self.element_components):
            onset_energies[ii] = comp.onset_energy
        return onset_energies

    def estimate_autofit_powerlaw(self):
        """
        Estimates the indices used to get information on the starting position
        of the power law. This is a region before the first edge in the
        spectrum.
        """
        onset_energies = self._get_onset_energies()
        first_onset = np.min(onset_energies)
        ind_f = self.spectrum.get_energy_index(first_onset)
        ind_i = 5  # the 10 pixel is used as estimate

        if ind_f <= ind_i:
            print('powerlaw estimation region too small')
            return None
        indices = [ind_i, ind_f]
        return indices

    def visualize_autofit(self, bg):
        """
        Visualizes the region used for the estimation. Helpful for debugging
        """

        indices = self.estimate_autofit_powerlaw()

        if isinstance(self.spectrum, MultiSpectrum):
            s = self.spectrum.mean()
        else:
            s = self.spectrum

        xx = np.arange(bg.size)
        where = (xx >= indices[0]) & (xx <= indices[1])
        fig, ax = plt.subplots()
        ax.plot(s.energy_axis, s.data)
        ax.plot(bg.energy_axis, bg.data)
        ax.fill_between(bg.energy_axis, 0, s.data.max(),
                        where=where, color='green', alpha=0.5)

    def make_background(self):
        """
        Makes the background component.
        """
        spsh = self.spectrum.get_spectrumshape()
        if self.background_model == 'powerlaw':
            bg = PowerLaw(spsh, A=10, r=3)
            indices = self.estimate_autofit_powerlaw()

            if isinstance(self.spectrum, MultiSpectrum):
                s = self.spectrum.mean()
            else:
                s = self.spectrum

            bg.autofit(s, indices[0], indices[1])

            if self.feedback:
                self.visualize_autofit(bg)

        elif self.background_model == 'linear':
            rlist = np.linspace(1, 5, self.n_bgterms)
            bg = LinearBG(specshape=spsh, rlist=rlist)
        self.bg = bg

    def make_coreloss(self):
        """
        Makes the atomic cross sections components. It uses Zezhongs calculated
        GOS array to compute the cross sections.

        """
        spsh = self.spectrum.get_spectrumshape()

        # The components of the edges
        comp_elements = []
        for elem, edge, onset in zip(self.elements, self.edges, self.onsets):
            if self.gos_array == 'zhang':
                comp = ZezhongCoreLossEdgeCombined(spsh, 1, self.E0,
                                                   self.alpha, self.beta,
                                                   elem, edge, eshift=onset,
                                                   q_steps=self.qsteps)
            elif self.gos_array == 'kohl':
                comp = KohlLossEdgeCombined(spsh, 1, self.E0, self.alpha,
                                            self.beta, elem, edge,
                                            eshift=onset, q_steps=self.qsteps)

            comp_elements.append(comp)
        self.element_components = comp_elements

    def estimate_sampling(self):
        if self.is_multispectrum:
            self.dE = np.sqrt(2)*self.ll.mean().get_numerical_fwhm()
        else:
            self.dE = np.sqrt(2)*self.ll.get_numerical_fwhm()

    def make_finestructure(self):
        """
        Creates fine structure components for edge core loss edge. It uses
        the fine_intervals attribute to know the interval region of the
        fine structure. The sampling is determined by the fwhm of the zlp.
        The self.pre_fine is used to indicate how much before the edge
        onset the fine structure should start.
        """
        spsh = self.spectrum.get_spectrumshape()

        # if no dE is given then estimate it from zero loss peak
        if self.dE is None:
            self.estimate_sampling()

        print(self.dE)
        for ii, comp in enumerate(self.element_components):
            n = int(self.fine_intervals[ii]/self.dE)
            print(n)
            if self.pre_fine is None:
                pre_fine = 0
            else:
                pre_fine = self.pre_fine[ii]
            fine = GDOSLin.gdoslin_from_edge(spsh, comp,
                                             ewidth=self.fine_intervals[ii],
                                             degree=n,
                                             interpolationtype='cubic',
                                             pre_e=pre_fine)

            self.fine_components.append(fine)

    def make_mscatter(self):
        """
        Creates the multiple scattering component. It uses the lowloss
        for this.
        """

        spsh = self.spectrum.get_spectrumshape()

        if self.ll is None:
            self.llcomp = None
        else:
            self.llcomp = MscatterFFT(spsh, self.ll)

    def make_model(self):
        """
        Creates the model from the available components.
        """
        spsh = self.spectrum.get_spectrumshape()

        self.make_coreloss()
        self.make_background()
        self.make_mscatter()

        components = [self.bg] + self.element_components
        if self.use_fine:
            self.make_finestructure()
            components.extend(self.fine_components)
        if self.llcomp is not None:
            components.append(self.llcomp)

        self.model = Model(spsh, components)

    def make_fitter(self):
        """
        Makes the fitter object. If the self.background_model is linear then
        a linear fitter is used. Else a non-linear least squares fitter
        will be used.

        """
        if self.background_model == 'linear':
            self.fitter = LinearFitter(self.spectrum, self.model,
                                       method=self.linear_fitter_method,
                                       use_weights=self.use_weights)

        else:
            for comps in self.element_components:
                comps.parameters[0].setboundaries(0, 1e10)
            self.fitter = LSQFitter(self.spectrum, self.model, method='trf',
                                    use_bounds=True)

    def do_procedure(self):
        """
        Workflow to process core loss results. First it aligns the
        multispectrum using the low loss. Then the model is calculated and
        the fitter object is created and the fitting procedure begins.

        """

        if self.do_align:
            self.align_zlp()
        self.make_model()
        self.make_fitter()

        if isinstance(self.spectrum, MultiSpectrum):
            self.fitter.multi_fit()
            self.multimodel = self.fitter.model_to_multispectrum()
            fig, maps, names = \
                self.fitter.show_map_result(self.element_components)
            self.elemental_maps = maps
            self.elemental_names = names
        else:
            self.fitter.perform_fit()

    def show_elements_maps(self):
        """
        Shows the elemental maps.
        """
        fig, maps, names = self.fitter.show_map_result(self.element_components)

    def get_elemental_maps(self):
        maps, names = self.fitter.get_map_results(self.element_components)
        self.elemental_maps = maps
        return maps, names

    def get_multimodels(self):
        """
        Calculates the different multimodels of the different components.
        The first multimodel is contains the fitted background. Then it adds
        the next edge until all edges are added. Hence the last multimodel in
        the list is the resulting fit. The inbetween results help interpreting
        the fitted cross sections.

        Returns
        -------
        multimodels: list of MultiSpectrum
            A list constaining multimodels to visualize how the resulting fit
            A list constaining multimodels to visualize how the resulting fit
            looks like.

        """
        multimodels = []
        multimodels.append(
            self.fitter.model_to_multispectrum_with_comps([self.bg]))

        onset_energies = self._get_onset_energies()
        co = np.argsort(onset_energies)

        use_comps = [self.bg]
        for ii in range(co.size):
            comp = self.element_components[co[ii]]
            comps = use_comps + [comp]
            s = self.fitter.model_to_multispectrum_with_comps(comps)
            multimodels.append(s)
            use_comps.append(comp)
            if self.use_fine:
                comp = self.fine_components[co[ii]]
                comps = use_comps + [comp]
                s = self.fitter.model_to_multispectrum_with_comps(comps)
                multimodels.append(s)
                use_comps.append(comp)

        return multimodels

    def get_CRLB(self):
        """
        Returns the CRLB for the different components of the coreloss
        edges.
        """
        if self.is_multispectrum:
            crlbs = np.zeros((len(self.element_components),
                              self.spectrum.xsize, self.spectrum.ysize))

        else:
            crlbs = np.zeros(len(self.element_components))

        for ii, comp in enumerate(self.element_components):
            if self.is_multispectrum:
                crlbs[ii] = self.fitter.CRLB_map(comp.parameters[0])
            else:
                crlbs[ii] = self.fitter.CRLB(comp.parameters[0])

        self.crlbs = crlbs
        return crlbs



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
import numpy as np
import matplotlib.pyplot as plt
import h5py
from os.path import exists
from scipy import interpolate
import os
from scipy.ndimage.filters import gaussian_filter

from pyEELSMODEL.io_tools.dm_ncempy import dmReader
from pyEELSMODEL.io_tools.hdf5_io import load_h5py
# from pyEELSMODEL import __file__  # when icon is used
# from PyQt5 import QtGui  # when icon is used

import copy
import logging

logger = logging.getLogger(__name__)


class Spectrumshape():
    """
    Spectrumshape is a class holding the main parameters of a spectrum in order
    to compactly create several spectra with the same size by providing an
    instance of this class
    It holds:
        dispersion: float, dispersion in eV/pixel describing the energy scale
        offset: float, energy of the first pixel in the spectrum in eV
        size: int, the size of a spectrum in number of pixels (eg typ 1024)
    """

    def __init__(self, dispersion, offset, size):
        self.dispersion = dispersion
        self.offset = offset
        self.size = size


class Spectrum:
    """
    Spectrum object which contains the experimental data.
    """

    def __init__(self, spectrumshape, data=None, acq_time=1, pppc=1):

        """
        Initialises a Spectrum instance

        Parameters
        ----------
        spectrumshape : Spectrumshape
            holds the basic parameters dispersion [eV/pixel], offset [eV] and
            size [pixels] to create the energy scale of the spectrum
        data : float or int, optional
            Holds the EELS intensity data with same size as spectrumshape.size.
            If None is given an empty spectrum is created with all zeros.
            (default is None).
        acq_time: float [s]
            The acquisition time for the spectrum. This attribute gives the
            ability to compare spectra with each when other acquisition
            times are used. (default: 1)
        pppc: primary particles per count, this is the value one electron
            generates.

        Raises
        ------
        ValueError
            if data does not have same size as in spectrumshape.size
        TypeError
            if data is neither int or float

        Returns
        -------
        An instance of a Spectrum.

        """
        self._dispersion = spectrumshape.dispersion
        self._offset = spectrumshape.offset
        self.size = spectrumshape.size
        self.eshift = 0
        self._set_energy_axis()

        if data is None:
            self.data = np.zeros(self.size)

        else:
            if len(data) != self.size:
                raise ValueError('data needs to be same size as spectrum.')
            if type(float(data[0])) != float:
                raise TypeError('data needs to be convertible to float.')
            self.data = data
        # Array which points to exclude
        self.exclude = np.zeros(self.size, dtype=bool)
        self.pppc = pppc
        self.name = 'a spectrum'
        self.acq_time = acq_time

        logger.debug('Spectrum init succeeded')

    @classmethod
    def load(cls, filename, flip_sign=False, **kwargs):
        """
        Loads different types of data. The possible datatypes are: .hdf5,
        .hspy, .dm3/.dm4 and .msa.

        Parameters
        ----------
        filename : string
           String of the filename where the extention should be '.hdf5'
        flip_sign: boolean
            Indicates whether the offset value should be negative when loading
            a .dm file. (default: False)

        Returns
        -------
        s: Spectrum
            The spectrum which is contained in the filename

        """
        ext = os.path.splitext(filename)[-1]

        if ext == '.hdf5':
            s = cls.load_hdf5(filename, **kwargs)

        elif ext == '.hspy':
            s = cls.load_hspy(filename)

        elif (ext == '.dm3') or (ext == '.dm4'):
            s = cls.load_dm(filename, flip_sign)

        elif ext == '.msa':
            s = cls.loadmsa(filename)

        else:
            raise ValueError(r'Extension is not valid')

        return s

    @classmethod
    def load_hdf5(cls, filename, return_metadata=False):
        """
        Loads the hdf5 file.

        Parameters
        ----------
        filename : string
           String of the filename where the extention should be '.hdf5'

        Returns
        -------
        s: Spectrum
            The spectrum which is contained in the filename

        """
        data, dispersion, offset, size = load_h5py(filename)

        specshape = Spectrumshape(dispersion, offset, size)
        s = Spectrum(specshape, data=data)
        if return_metadata:
            return s,
        else:
            return s

    @classmethod
    def load_hspy(cls):
        pass

    @classmethod
    def from_numpy(cls, data_array, energy_axis):
        """
        Creates a Spectrum object from the data and energy axis using
        numpy arrays. Usefull when playing with different functions which are
        not (yet) integrated into pyEELSmodel functionalities.

        Parameters
        ----------
        data_array : numpy array (1D)
            The data of the EEL spectrum.
        energy_axis: numpy array (1D)
            The energy axis used [eV]

        Returns
        ----------
        s: Spectrum

        """
        dispersion = energy_axis[1] - energy_axis[0]
        offset = energy_axis[0]
        size = energy_axis.size
        specshape = Spectrumshape(dispersion, offset, size)
        s = Spectrum(specshape, data=data_array)
        return s

    def _set_energy_axis(self):
        """
        Prepares the energy axis
        Returns
        -------
        None.
        """
        self._energy_axis = np.arange(self.size, dtype='float') * \
            self._dispersion + self._offset + self.eshift

    @property
    def energy_axis(self):
        return self._energy_axis

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, offset):
        self._offset = offset
        self._set_energy_axis()

    @property
    def dispersion(self):
        return self._dispersion

    @dispersion.setter
    def dispersion(self, dispersion):
        self._dispersion = dispersion
        self._set_energy_axis()

    @property
    def acq_time(self):
        """
        The acquisition time used for spectra. Can be usefull when splicing
        multiple spectra together.

        """
        return self._acq_time

    @acq_time.setter
    def acq_time(self, acq_time):
        self._acq_time = acq_time

    def setname(self, name):
        """
        Set the name of a spectrum. Can be used when displaying the spectrum.
        Could hold e.g. the filename if loaded from a file.
        """
        self.name = name

    def copy(self):
        """
        Returns a copy of the object
        """
        return copy.deepcopy(self)

    @property
    def eshift(self):
        return self._eshift

    @eshift.setter
    def eshift(self, eshift):
        self._eshift = eshift
        self._set_energy_axis()

    def __mul__(self, other):
        """
        Multiply two spectra with each other. All properties of the self
        spectrum are copied, except the data is multiplied with the data of the
        other spectrum. If other is an int or float, the spectrum is multiplied
        by that number.

        Parameters
        ----------
        other : Spectrum
            The spectrum to multiply with.

        Raises
        ------
        ValueError
            In case both spectra don't have the same spectrumshape settings
            (in which case multiply makes no sense).
        TypeError
            In case other is neither a Spectrum, int or float.

        Returns
        -------
        s : Spectrum
            Returns a reference to a -new- spectrum that is the multiplication
            result.

        """
        if isinstance(other, Spectrum):
            if self._check_not_same_settings(other):
                raise ValueError('Dispersion, offet or size are not the same')
            else:
                s = self.copy()
                s.data = self.data * other.data
                return s

        elif type(other) is int or type(other) is float:
            s = self.copy()
            s.data = self.data * other
            return s

        else:
            raise TypeError('Input should be spectrum object, float or int')

    def __add__(self, other):
        """
        Add two spectra. All properties of the self spectrum are
        copied, except the data is added with the data of the other spectrum.
        If other is an int or float, the number is added to that spectrum.

        Parameters
        ----------
        other : Spectrum
            The spectrum to add.

        Raises
        ------
        ValueError
            In case both spectra don't have the same spectrumshape settings
            (in which case adding makes no sense).
        TypeError
            In case other is neither a Spectrum, int or float.

        Returns
        -------
        s : Spectrum
            Returns a reference to a -new- spectrum that is the addition
            result.

        """
        if isinstance(other, Spectrum):
            if self._check_not_same_settings(other):
                raise ValueError('Disperion, offet or size are not the same')

            else:
                s = self.copy()
                s.data = self.data + other.data
                return s

        elif type(other) is int or type(other) is float:
            s = self.copy()
            s.data = self.data + other
            return s

        else:
            raise TypeError('Input should be spectrum object, float or int')

    def __sub__(self, other):
        """
        Subtract two spectra. All properties of the self spectrum are
        copied, except the data is added with the data of the other spectrum.
        If other is an int or float, the number is subtracted from that
        spectrum.

        Parameters
        ----------
        other : Spectrum
            The spectrum to subtract.

        Raises
        ------
        ValueError
            In case both spectra don't have the same spectrumshape settings
            (in which case subtracting makes no sense).
        TypeError
            In case other is neither a Spectrum, int or float.

        Returns
        -------
        s : Spectrum
            Returns a reference to a -new- spectrum that is the subtraction
            result.

        """
        if isinstance(other, Spectrum):
            if self._check_not_same_settings(other):
                raise ValueError('Dispersion, offet or size are not the same')

            else:
                s = self.copy()
                s.data = self.data - other.data
                return s
        elif type(other) is int or type(other) is float:
            s = self.copy()
            s.data = self.data - other
            return s

        else:
            raise TypeError('Input should be spectrum object, float or int')

    def __truediv__(self, other):
        return self.__div__(other)

    def __div__(self, other):
        """
        Divide two spectra by each other. All properties of the self spectrum
        are copied, except the data is divided with the data of the other
        spectrum. If other is an int or float, the spectrum is divided by that
        number.

        Parameters
        ----------
        other : Spectrum
            The spectrum to divide by.

        Raises
        ------
        ValueError
            In case both spectra don't have the same spectrumshape settings
            (in which case divide makes no sense).
        TypeError
            In case other is neither a Spectrum, int or float.

        Returns
        -------
        s : Spectrum
            Returns a reference to a -new- spectrum that is the division
            result.

        """
        if isinstance(other, Spectrum):
            if self._check_not_same_settings(other):
                raise ValueError('Dispersion, offset or size are not the same')

            else:
                s = self.copy()
                s.data = self.data / other.data
                return s

        elif type(other) is int or type(other) is float:
            s = self.copy()
            s.data = self.data / other
            return s

        else:
            raise TypeError('Input should be spectrum object, float or int')

    def __getitem__(self, key):
        """
        [] operator, makes a copy of the spectrum where key
        is the indexing used. Slicing the energy axis is not possible and
        the get_interval method should be used.


        Parameters
        ----------
        key : indexing
            Indexing of the multispectrum can be performed as is done with
            numpy arrays.

        Raises
        ------
        ValueError
            DESCRIPTION.

        Returns
        -------
        s: MultiSpectrum/Spectrum
            The sliced multispectrum or spectrum

        #TODO add some warining when slicing the energy loss spectrum
        """

        # check if the energy axis is sliced
        print(type(key))
        if type(key) is dict:
            if key.start is None:
                ind0 = 0
            else:
                ind0 = key.start
        elif type(key) is int:
            ind0 = key
        elif type(key) is slice:
            if key.start is None:
                ind0 = 0
            else:
                ind0 = key.start
        else:
            raise TypeError('__getitem__ key should be dictionary or int or '
                            'slice')
        ndata = self.data[key]
        noffset = self.offset + ind0 * self.dispersion
        sh = Spectrumshape(self.dispersion, noffset, ndata.size)
        s = Spectrum(sh, data=ndata)
        return s

    def get_interval(self, interval):
        """
        Returns a subspectrum of the given interval.

        Parameters
        ----------
        interval : tuple
            Interval which has the start and end of the sub spectrum [eV]

        Returns
        -------
        s : Spectrum
            Sub spectrum given by the interval.

        """

        ind0 = self.get_energy_index(interval[0])
        ind1 = self.get_energy_index(interval[1])
        s = self[ind0:ind1]
        return s

    def _check_not_same_settings(self, other):
        """
        Checks if the other spectrum has the same spectrumshape properties as
        self.

        Parameters
        ----------
        other : Spectrum
            A spectrum to compare with.

        Returns
        -------
        bool
            True if both spectra have the same spectrumshape, False otherwise.

        """
        bl0 = self.dispersion != other.dispersion
        bl1 = self.offset != other.offset
        bl2 = self.size != other.size

        return bl0 or bl1 or bl2

    def get_specparameters(self):
        """
        Returns a tuple containing dispersion, offset, size of spectrum.
        Consider using get_spectrumshape() instead which returns a class
        holding the same information but in a way it can more easily be used to
        construct a new spectrum with same basic properties

        Returns
        -------
        dispersion : float
            Energy dispersion in eV/pixel.
        offset : float
            Energy of the first bin in the spectrum in eV.
        size : int
            Number of spectral bins.

        """
        return (self.dispersion, self.offset, self.size)

    def get_spectrumshape(self):
        """
        Return a Spectrumshape object holding information on the dispersion,
        offset, size of the Spectrum.

        Returns
        -------
        shape : Spectrumshape
            A spectrumshape object.

        """
        shape = Spectrumshape(self.dispersion, self.offset, self.size)
        return shape

    def bad_index(self, index: int):
        """
        Check if index is in range otherwise raises a ValueError

        Parameters
        ----------
        index : int
            Index from 0..size into the spectral data.

        Raises
        ------
        ValueError
            If index <0 or index>self.size.

        Returns
        -------
        None.

        """
        if index < -1 or index > self.size:
            raise ValueError('Bad index value')

    def set_exclude_region_energy(self, Ei, Ef):
        """
        Sets the excluded region using the energy.

        Parameters
        ----------
        Ei: float
            The starting energy of the excluded region.
        Ef: float
            The end energy of the excluded region.

        """
        index_i = self.get_energy_index(Ei)
        index_f = self.get_energy_index(Ef)
        self.set_exclude_region(index_i, index_f)

    def set_include_region_energy(self, Ei, Ef):
        """
        Sets the included region using the energy.

        Parameters
        ----------
        Ei: float
            The starting energy of the excluded region.
        Ef: float
            The end energy of the excluded region.

        """
        index_i = self.get_energy_index(Ei)
        index_f = self.get_energy_index(Ef)
        self.set_include_region(index_i, index_f)

    def set_exclude_region(self, index_i, index_f):
        """
        Set a range of pixels from index_i to index_f (both inclusive) as
        'excluded'. This means the Fitter will not take into account these
        spectral regions. Note that you can call this function multiple times
        on different index ranges to define multiple areas of exclusion.

        Parameters
        ----------
        index_i: int
            starting index for the exclusion area.
        index_f: int
            end index for the exclusion area.

        Returns
        -------
        None.

        """
        self.bad_index(index_i)
        # if the start or end index are not inside the region, throw error
        self.bad_index(index_f)
        self.exclude[index_i:index_f] = True

    def set_include_region(self, index_i, index_f):
        """
        Set a range of pixels from index_i to index_f (both inclusive) as
        'included'. This means the Fitter will take into account these spectral
        regions. Note that you can call this function multiple times on
        different index ranges to define multiple areas of exclusion.

        Parameters
        ----------
        index_i : int
            starting index for the inclusion area.
        index_f : int
            end index for the inclusion area.

        Returns
        -------
        None.

        """
        self.bad_index(index_i)
        # if the start or end index are not inside the region, throw error
        self.bad_index(index_f)
        self.exclude[index_i:index_f] = False

    def reset_exclude_region(self, index_i=0, index_f=-1):
        """
        Reset a range of pixels from index_i to index_f (both inclusive) from
        being 'excluded'. This means the Fitter will take into account these
        spectral regions. Note that you can call this function multiple times
        on different index ranges to define multiple areas as being not
        excluded. Note that default behaviour is that all spectral bins are not
        excluded when creating a spectrum.

        Parameters
        ----------
        index_i : int
            starting index for the inclusion area.
        index_f : int
            end index for the inclusion area.

        Returns
        -------
        None.

        """
        self.bad_index(index_i)
        # if the start or end index are not inside the region, throw error
        self.bad_index(index_f)
        self.exclude[index_i:index_f] = False

    def getmaxindex(self):
        """
        Returns the index containing the highest value in the spectrum. In case
        of multiple pixels having the same maximum value, index to the first
        occurence will be returned.
        """
        return np.argmax(self.data)

    def getmaxenergy(self):
        co = self.getmaxindex()
        return self.energy_axis[co]

    def get_first_higher_then(self, x):
        """
        Get index of the first pixel having a value higher then x

        Parameters
        ----------
        x : float
            The value of which the pixel value should be larger.

        Returns
        -------
        int
            index of the first occurence of data>x.
        None
            if no value higher than x is found

        """
        # todo add type checking of x?

        boolean = self.data > x
        if boolean.any():
            return np.argmax(boolean)
        else:
            logger.warning('No larger value then %d in the data', x)
            return None

    def get_max(self):
        """
        Returns the maximum intensity value in the spectrum.

        Returns
        -------
        float or int
            Maximum value in the EELS dataset.

        """
        return self.data.max()

    def get_min(self):
        """
        Returns the minimum intensity value in the spectrum.

        Returns
        -------
        float or int
            Minimum value in the EELS dataset.

        """
        return self.data.min()

    @property
    def pppc(self):
        """
        Get 'primary particles per count=pppc'. This value defines the number
        of electrons per given EELS count and is used to include information on
        the detector step. This is important for scintillating detectors where
        multiple counts can be generated for a single incoming electron.
        Knowing this is essential to calculate the effect of Poisson noise for
        the fitter. This only makes sense for experimental data and should not
        be used in e.g. components.

        Returns
        -------
        float
            The pppc value.

        """

        return self._pppc

    @pppc.setter
    def pppc(self, p):
        """
        Set the 'primary particles per count=pppc' value. This value defines
        the number of electrons per given EELS count and is used to include
        information on the detector step. This is important for scintillating
        detectors where multiple counts can be generated for a single incoming
        electron. Based on this, the estimated standard deviation is
        pre-calculated for the data which is used by the fitter. For Component
        objects, the default is pppc=1 and should not be set.

        Parameters
        ----------
        p : float
            the pppc value obtained e.g. from a test script determining the
            noise properties of the detector.


        Raises
        ------
        ValueError
            when pppc <0

        Returns
        -------
        None.

        """
        if p < 0:
            raise ValueError('pppc should be >=0')
        self._pppc = p
        # self.data = self.data*p
        self.init_poisson_error()

    def init_poisson_error(self):
        """
        Calculate estimated standard deviation on the data assuming underlying
        Poisson statistics. Result can be obtained with get_error(). Any
        non-positive data is ignored and result in a standard deviation of 0

        Returns
        -------
        None.

        """
        mask = self.data > 0
        self.error = np.sqrt(
            self.pppc * self.data * mask)  # only makes sense for positive data

    def get_energy_index(self, E):
        """
        Return the first index where E>energy within the spectral range

        Parameters
        ----------
        E : float
            Energy to find in this spectrum [eV].

        Returns
        -------
        int
            Index of the closest energy bin.
            0 when below the energy onset of the spectrum
            self.size when above the energy range of the spectrum

        """
        if E < self.energy_axis[0]:
            return 0
        elif E >= self.energy_axis[-1]:
            return self.size
        else:
            co = np.argmin(np.abs(self.energy_axis - E))
            return co

    def plot(self, externalplt=None, use_e_axis=True, logscale=False,
             **kwargs):
        """
        Plot the spectrum

        Parameters
        ----------
        externalplt : matplotlib reference
              A reference to an external matplotlib reference, if None we use
              our own matplotlib and create a new figure.
        use_e_axis: bool
            Indicates if the x-axis in the energy axis or if the pixel value
            of the detector will be used to visualize the data.

        """
        tempplt = plt
        if isinstance(externalplt, plt.Figure):
            tempplt = externalplt
        else:
            # create our own figure
            plt.figure()
            # thismanager = plt.get_current_fig_manager()

            # dirname= os.path.dirname(__file__) + "/../pyEELSMODEL/images/"
            # icon_name = os.path.join(dirname, 'test_logo.ico')
            # thismanager.window.wm_iconbitmap(icon_name)
            # thismanager.window.setWindowIcon(QtGui.QIcon(icon_name))

        if use_e_axis:
            tempplt.plot(self.energy_axis, self.data, **kwargs)
            tempplt.xlabel(r'Energy Loss [eV]')
        else:
            tempplt.plot(self.data, **kwargs)

        tempplt.ylabel('Counts')
        if logscale:
            tempplt.yscale('log')

    def show_excluded_region(self, **kwargs):
        """
        Shows the points which are excluded from the fit.
        """

        fig, ax = plt.subplots()
        ax.plot(self.energy_axis, self.data, color='black')
        ax.fill_between(self.energy_axis, 0, self.data.max(),
                        where=self.exclude, color='green', alpha=0.5)

    def erase(self):
        self.data = np.zeros(self.size)

    def normalise(self, method='sum', window=None):
        """
        Normalizes the spectrum using the method identified. The data is over-
        written

        Parameters
        ----------
        method : string
            The normalization method which can be 'sum' or 'max'.
            (default: 'sum')
        window: tuple
            If the method is 'sum', an integration range can be added. If
            no window is provided the sum under the entire spectrum is taken.

        """
        if method == 'max':
            self.data = self.data / self.data.max()

        elif method == 'sum':
            if window is None:
                norm = np.sum(self.data)
            else:
                norm = self.integrate(window)
            self.data = self.data / norm

        else:
            print('method is not supported')

    def _getmsatag(self, line, tag):
        """
        Look for 'tag' in string 'line' and convert value after ':' to float
        return, return -1 if tag not found, return 0 if tag found but value
        couldnt be interpreted

        Parameters
        ----------
        line : TYPE
            DESCRIPTION.
        tag : TYPE
            DESCRIPTION.
        ref : TYPE
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        val = 0
        # logger.debug('getmsatag on line=%s and looking for tag=%s',line,tag)
        if line.find(tag) != -1:
            logger.debug('item found %s', tag)
            items = line.split(':')
            try:
                val = float(items[-1])
            except (ValueError, IndexError):
                # if value cant be converted we still want True as
                # this is the case for #SPECTRUM which has no value
                val = 0
            return val
        return -1

    @classmethod
    def loadmsa(cls, filename):
        """
        Load single spectrum from an MSA file

        Parameters
        ----------
        filename : string
            The name of the .msa which needs to opened.

        Returns
        -------
        s: Spectrum
            The EEL spectrum in the file

        """
        logger.warning('this will change the size and contents of the current '
                       'spectrum')
        specdata = False
        i = 0
        specshape = Spectrumshape(1, 200, 1024)
        s = Spectrum(specshape)
        with open(filename, 'r') as f:
            for line in f:
                if not specdata:
                    val = s._getmsatag(line, '#NPOINTS')
                    if val != -1:
                        s.size = int(val)
                        s.erase()
                    val = s._getmsatag(line, '#XPERCHAN')
                    if val != -1:
                        s.dispersion = val
                    val = s._getmsatag(line, '#OFFSET')
                    if val != -1:
                        s.offset = val
                    if s._getmsatag(line, '#SPECTRUM') != -1:
                        specdata = True
                else:
                    items = line.split(',')
                    try:
                        first = float(items[0])
                    except (ValueError, IndexError):
                        pass
                    try:
                        second = float(items[-1])
                    except (ValueError, IndexError):
                        pass
                    logger.debug('reading data: %s,%s', first, second)
                    if i < s.size:
                        s.data[i] = float(second)
                    i += 1
        logger.info('MSA file loaded')
        return s

    @classmethod
    def load_dm(cls, filename, flip_sign=False):
        """
        Load single spectrum from an DM file. There seems to be a bug in the
        dm loader which gives a positive offset value while it should be
        negative. This happens mainly for the low loss spectra which contain
        the zero-loss peak. The flip_sign parameter takes care of this.

        Parameters
        ----------
        filename : string
            filename of the dm file
        flip_sign: boolean
            Indicates where the offset value should be set to negative.
            (default: False)

        Returns
        -------
        s: Spectrum

        """
        dmfile = dmReader(filename)
        data = dmfile['data'][:]
        e_axis = dmfile['coords'][0]
        dispersion = e_axis[1] - e_axis[0]
        offset = e_axis[0]
        if flip_sign:
            offset = -1 * offset
        size = e_axis.size
        spc = Spectrumshape(dispersion, offset, size)
        return Spectrum(spc, data)

    def save_hdf5(self, filename, metadata=None, overwrite=False):
        """
        Saves the spectrum as a hdf5 file. The structure of the file can easily
        be investigated via a hdfview software.

        Parameters
        ----------
        filename : string
            filename of the saved file.
        metadata: dictionary
            A dictionary containing E0, alpha, beta, elements and edges can be
            added to the hdf5 file. If None is given, nothing will be saved.
        overwrite: boolean
            Indicates if the file will be overwritten if it already exists.
            (default: False)
        Returns
        -------
        If the file saving workes, a True value is returned.
        """
        file_exists = exists(filename)
        if not overwrite and file_exists:
            logger.info(r'File already exists, set overwrite to True if you '
                        r'want to overwrite existing file')
            return False
        if file_exists:
            os.remove(filename)

        f = h5py.File(filename, 'w')
        f.create_dataset('Data', data=self.data)
        d = f.create_group('Energy_axis')

        d.attrs['Dispersion'] = self.dispersion
        d.attrs['Offset'] = self.offset
        d.attrs['Size'] = self.size

        if metadata is not None:
            d.attrs['alpha'] = metadata['alpha']
            d.attrs['beta'] = metadata['beta']
            d.attrs['E0'] = metadata['E0']
            d.attrs['elements'] = metadata['elements']
            d.attrs['edges'] = metadata['edges']

        f.close()
        return True

    def gaussiansmooth(self, sigma):
        """
        Return a new spectrum which is a gaussian smoothed version of this one.

        Parameters
        ----------
        sigma : float
            Standard deviation of the gaussian with which it will be
            convolved


        Returns
        -------
        result : Spectrum
            A smoothed spectrum version of the initial spectrum.

        """
        result = self.copy()
        result.data = gaussian_filter(self.data, sigma / self.dispersion)
        return result

    def getnonexcludedpoints(self):
        """
        Returns the number of included points in the spectrum, used for the
        degrees of freedom.
        """
        return np.size(self.exclude) - np.count_nonzero(self.exclude)

    def integrate(self, window=None, index_type=False):
        # todo add test for this function
        # todo add badindex check
        """
        Integrates the signal over the given energy window or the given index.

        Parameters
        ----------
        window : tuple
            The window over which to integrate. If index_type is True,
            the integration window is given with the index instead of energy
            If window is None, the integration is performed over the entire
            energy range.
        index_type: bool
            Indicates if the window should be interpreted as the indices or
            the energy of the spectrum

        Returns
        -------
        result : float
            The value of the integration

        """
        if window is None:
            window = [0, self.size]
            index_type = True

        if index_type:
            result = self.data[window[0]:window[1]].sum()
        else:
            ind0 = self.get_energy_index(window[0])
            ind1 = self.get_energy_index(window[1])
            result = self.data[ind0:ind1].sum()
        return result

    def interp_to_other_energy_axis(self, spectrum, constant_values=(0, 0)):
        """
        Interpolates the given spectrum to another given spectrum. The region
        which does not overlap with the energy axis of spectrum will be given a
        zero value. Other methods of padding can easily be incorporated using
        numpys pad functionality. The interpolation is linear and can also be
        modified using the interp function of scipy.

        Parameters
        ----------
        spectrum : Spectrum
            The spectrum which holds energy axis to which the spectrum needs to
             be interpolated.

        Returns
        -------
        int_sepc : Spectrum
            A new spectrum which is interpolated to the given spectrums energy
            axis.
        """

        disp = self.dispersion
        dstart = -1 * min(spectrum.offset - self.offset, 0)
        dend = max(spectrum.energy_axis[-1] - self.energy_axis[-1], 0)

        # elongate the energy axis such the entire area can be interpolated.
        if dstart == 0:
            before = []
        else:
            before = np.arange(np.ceil(dstart / disp)) * disp
            before += self.offset - before[-1] - disp

        if dend == 0:
            after = []
        else:
            after = np.arange(np.ceil(dend / disp)) * disp \
                    + self.energy_axis[-1] + disp

        E_ax = np.concatenate((before, self.energy_axis, after))
        ndata = np.pad(self.data, pad_width=(len(before), len(after)),
                       constant_values=constant_values)
        f = interpolate.interp1d(E_ax, ndata)
        int_spec = Spectrum(spectrum.get_spectrumshape(),
                            data=f(spectrum.energy_axis))
        return int_spec

    def rescale_spectrum(self, scale, shift):
        """
        Modifies the EELS spectrum by keeping the energy axis the same but
        shifting and rescaling the spectrum. First the shift is applied
        and then the scaling is modified.

        Parameters
        ----------
        scale : float
            The value with which the dispersion is multiplied.

        shift: float [eV]
            The shift of the spectrum on the energy axis.

        Returns
        -------
        int_spec : Spectrum
            A new spectrum has been shifted and rescaled but has the same
            energy axis.
        """
        cop = self.copy()
        cop.offset += shift
        cop.dispersion = scale * self.dispersion
        int_spec = cop.interp_to_other_energy_axis(self)
        return int_spec

    def padding(self, padding):
        """
        Zero padding the spectrum with padding size. This function is used
        when convolving the zero loss spectrum.

        Parameters
        ----------
        padding : uint
            The number of elements padded on the left and right

        Returns
        -------
        s : Spectrum object
            A new spectrum which is zero padded.
        """
        ndata = np.pad(self.data, (padding, padding))
        n_offset = self.offset - padding * self.dispersion
        hs = Spectrumshape(self.dispersion, n_offset, ndata.size)
        s = Spectrum(hs, data=ndata)
        s.set_exclude_region(0, padding + 1)
        s.set_exclude_region(self.size + padding - 1, s.size)
        return s

    def recalibrate_spectrum(self, coefficients):
        """
        Recalibrate the spectrum using the coefficient for changing the
        dispersion and offset. These coefficients can be retrieved by linear
        fitting the measured edge onset energies compared to the literature
        values.

        Parameters
        ----------
        coefficients : numpy array 1D
            This array has two values coming from a linear fit (y = ax + b).
            The first element is the slope of the fit (a) and the second
            element is the constant (b).

        """
        n_en_axis = coefficients[0] * self.energy_axis + coefficients[1]
        disp = n_en_axis[1] - n_en_axis[0]
        offset = n_en_axis[0]
        self.dispersion = disp
        self.offset = offset

    def firstderivative(self):
        """
        Calculates the second derivative of the spectrum.
        Returns
        -------
        s: Spectrum
            The first derivative for the spectrum
        """

        differential = np.zeros_like(self.data)
        differential[0] = self.data[0] - 2 * self.data[1]
        differential[-1] = self.data[-1] - 2 * self.data[-2]
        differential[1:-1] = self.data[0:-2] - 2 * self.data[1:-1] \
            + self.data[2:]

        nspec = Spectrumshape(self.dispersion, self.offset, self.size)
        s = Spectrum(nspec, data=differential)
        return s

    def secondderivative(self):
        """
        Calculates the second derivative of the spectrum.

        Returns
        -------
        s: Spectrum
            The second derivative for the spectrum
        """
        differential = np.zeros_like(self.data)
        differential[0] = self.data[0] - 2. * self.data[1] + self.data[2]
        differential[-1] = self.data[-1] - 2. * self.data[-2] + self.data[-3]
        differential[1:-1] = self.data[0:-2] - 2. * self.data[1:-1] \
            + self.data[2:]

        differential /= self.dispersion ** 2

        nspec = Spectrumshape(self.dispersion, self.offset, self.size)
        s = Spectrum(nspec, data=differential)
        return s

    def apply_dark_and_gain(self, gain_name, dark_name):
        """
        Apply dark and gain. The dark and gain can be saved as a numpy array
        and applied to the spectrum. This is done on the data of the spectrum
        itself and no new Spectrum object is created.
        It assumes that the gain and dark is stored as a .npy file

        Parameters
        ----------
        gain_name : string
            The filename where the gain is stored.

        dark_name: string
            The filename where the dark reference is stored

        """
        self.apply_dark(dark_name)
        gain = np.load(gain_name, allow_pickle=True)
        self.data = gain * self.data

    def apply_dark(self, dark_name):
        """
        Remove the dark reference from the detector. This is done on the data
        of the spectrum itself and no new Spectrum object is created.


        Parameters
        ----------
        dark_name: string
            The filename where the dark reference is stored

        """
        dark = np.load(dark_name, allow_pickle=True)
        self.data -= dark

    def get_numerical_fwhm(self):
        """
        Small function which calculates the fwhm of a spectrum numerically
        by subtracting halve of the maximum from the spectrum and finding the
        smallest values. Note that the zero loss peak should be present else
         the result will be garbage.

        Returns
        -------
        fwhm: float
            The full width halve maximum [eV], None is returned if the
            offset energy of the spectrum is bigger or equal to zero.

        """
        if self.offset >= 0:
            print('The energy axis is positive indicating that the zero loss'
                  'peak is not present')
            return None

        rs = np.abs(self.data - 0.5 * self.data.max())
        co = np.argsort(rs)
        fwhm = np.abs(self.energy_axis[co[1]] - self.energy_axis[co[0]])
        return fwhm



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
import numpy as np
import matplotlib.pyplot as plt

from pyEELSMODEL.core.spectrum import Spectrum
from pyEELSMODEL.core.component import Component
from pyEELSMODEL.components.MScatter.mscatter import Mscatter
import logging

logger = logging.getLogger(__name__)


class Model(Spectrum):
    """
    Model class
    A Model is a Spectrum that can calculate itself based on a number of
    Components which each depend on a number of Parameters.
    Note that all actions like convolving with low loss, adding or multiplying
    stuff is encoded in the Components, some of which are special types which
    tells the model how to treat them to come up with a single model result.
    Philosophy of Model: a model with a number of parameters mimicking an EELS
    experiment. Everything that happens in experiments needs to be modelled
    somehow by the components in the model. The model doesnt care whether the
    spectrum it simulates is a single spectrum or an SI. It is the task of the
    fitter to deal with the parameters for each position in an SI
    (different from cpp EELSMODEL!)
    """

    def __init__(self, spectrumshape, components=None):

        """
        Instantiate a Model object.

        Parameters
        ----------
        spectrumshape : Spectrumshape
            Shape of the spectrum the model should mimick (see Spectrum
            documentation).
        components : list of Components, optional
            List of Components the model is made of, allows to load the model
            from a file. The default is None in which case the model will be
            empty to start.

        Returns
        -------
        None.

        """
        super().__init__(spectrumshape)
        if components is None:
            self.components = []
        else:
            self.components = components

        self.allparameters = []
        self.freeparameters = []
        self.freelinparameter = []
        self.freenonlinparameters = []
        self.locked = False
        self.changed = True
        self.prepare()

    def release(self):
        for comp in self.components:
            comp.release()

    def lock(self, b):
        """
        Set the locked state of the model. If the model is locked, no
        components can be added or removed. This is useful during fitting
        where the model structure should not change.

        Parameters
        ----------
        b : bool
            True=locked, False is unlocked.

        Returns
        -------
        None.

        """
        self.locked = b

    def islocked(self):
        return self.locked

    def setchanged(self, b):
        self.changed = b

    def ischanged(self):
        return self.changed

    def addcomponent(self, component):
        if self.islocked():
            # don't do it when the model is locked, the fitter typically locks
            # the model during fitting
            return
        if isinstance(component, Component):
            # give every new component the same eshift as we have
            # component.seteshift(self.geteshift())
            component.eshift = self.eshift
            self.components.append(component)
            self.setchanged(True)
            self.prepare()
            self.calculate()

        else:
            raise TypeError(r'Input should be Component object')

    def removecomponent(self, component):
        """
        Removes a component from the model.

        Parameters
        ----------
        component : Component
            The component which needs to be removed.

        Returns
        -------
        None.

        """
        if self.islocked():
            return
        if isinstance(component, Component):
            self.components.remove(component)
            self.setchanged(True)
            self.prepare()
            self.calculate()
            return True
        else:
            raise TypeError(r'Input should be Component object')

    def getcomponents(self):
        return self.components

    def getallparameters(self):
        return self.allparameters

    def getfreeparameters(self):
        return self.freeparameters

    def getfreelinparameters(self):
        return self.freelinparameter

    def getfreenonlinparameters(self):
        return self.freenonlinparameters

    def prepare(self):
        """
        Reallocate space to hold parameters. Needs to be called every time the
        number of components or state of parameters (changeable or linear)
        changes. The parameterlists are what is handled by the fitter. So good
        time to call this function is right before a fitting loop to make sure
        we are working with an up to date status of all parameters. Changing
        parameter status (changeble or linear) during a fit should not be
        allowed in the UI. Calling this everytime before calculate would be
        too time consuming.

        Returns
        -------
        None.

        """
        # empty storage space
        self.allparameters = []
        self.freeparameters = []
        self.freelinparameter = []
        self.freenonlinparameters = []
        # and fill it up again
        for comp in self.components:
            for p in comp.parameters:
                self.allparameters.append(p)
                if p.ischangeable():
                    self.freeparameters.append(p)
                    if p.islinear():
                        self.freelinparameter.append(p)
                    else:
                        self.freenonlinparameters.append(p)

    def getcomponentbyparameter(self, parameter):
        """
        Get component that belongs to a given parameter reference.

        Parameters
        ----------
        parameter : Parameter
            Reference to a parameter.

        Returns
        -------
        comp : Component
            reference to a component that has this parameter reference.
            None if parameter not found in any component in the model.

        """
        for comp in self.components:
            if parameter in comp.parameters:
                return comp
        return None

    def islinear(self):
        """
        Returns whether all free parameters in the model are linear. If this
        is the case we have a full linear model which allows the use of a
        linear fitter.

        Returns
        -------
        boolean : bool
            True if linear model, False otherwise.

        """
        boolean = True
        for comp in self.components:
            for param in comp.parameters:
                if param.ischangeable():
                    boolean = boolean and param.islinear()
                    if not param.islinear():
                        logger.warning('non linear parameter found %s in '
                                       'component:%s', param.getname(),
                                       comp.getname())
        return boolean

    def getnumcomponents(self):
        """
        Return the number of components in the model.

        Returns
        -------
        int
            Number of components in the model.

        """
        self.prepare()  # make sure to return up to date information
        return len(self.components)

    def getnumparameters(self):
        """
        Get total number of parameters in the model. This includes the free
        and the non-free parameters.

        Returns
        -------
        n_param : int
            Number of parameters in the model.

        """
        self.prepare()  # make sure to return up to date information
        return len(self.allparameters)

    def getnumfreeparameters(self):
        """
        Get total number of free parameters in the model.

        Returns
        -------
        n_free : int
            Number of free parameters in the model.

        """
        self.prepare()  # make sure to return up to date information
        return len(self.freeparameters)

    def saveparams(self, filehandle, fmt='txt', header=True):
        """
        Store all current parameters in a file that is already open
        this allows to append the parameters of all estimated params in a
        spectrum image in one big file.
        TODO work eg with elementtree for XML writing

        Parameters
        ----------
        filename : TYPE
            DESCRIPTION.
        fmt : TYPE, optional
            DESCRIPTION. The default is 'txt'.

        Returns
        -------
        None.

        """

        # in first line of text file should appear the names
        # if header:
        #     for p in allparameters:
        #         print(p.getdisplayname(),'\tCRLB\t',end='')
        #     print('')
        # #decide on text format style, tab separated is easy
        # for p in allparameters:
        #     print(p.getvalue(),'\t',p.getsigma,'\t',end='')
        print('')
        # replace with file writing
        # and more compact w binary

    def savemodel(self):
        print('saves the model')

    def order_coupled_components(self):
        """
        The components which have parameters which are coupled to other
        parameters should be first calculated in the model. Therefore the
        component list should be ordered such that first the coupled parameters
        are calculated and then the rest.
        """
        worker = []
        main = []

        for comp in self.components:
            b = False
            for param in comp.parameters:
                if param.iscoupled():
                    cm = self.getcomponentbyparameter(param.coupled_parameter)
                    if cm is not comp:
                        b = True

            if b:
                worker.append(comp)
            else:
                main.append(comp)

        components = worker + main
        self.components = components

    def calculate(self, use_ll=True):
        """
        Calculate the model by calculating all components.
        For the linear least squares fitting it would be interesting to not use
        the low loss for computing the model since we can do the low loss
        convolution at a later step to improve speed for the linear fitting.

        Parameters
        ----------
        use_ll: boolean
            Indicates if it needs to use the low loss for calculating the
            model.

        Raises
        ------
        ValueError
            DESCRIPTION.

        Returns
        -------
        None.

        """

        self.erase()  # start with a clean slate

        # apply shift components if present they come first since they change
        # the energy scale for all the rest
        for comp in self.components:
            if comp.getshifter():
                comp.calculate()

        # prepare multiplier components if present
        for comp in self.components:
            if comp.get_ismultiplier():
                comp.calculate()

        # calculate each component and add up
        # first all components which need to be convoluted

        # rearrange the order of the components such that the coupled ones
        # are calculated first
        self.order_coupled_components()

        for comp in self.components:
            if not comp._isconvolutor:
                if comp.getcanconvolute() and (not (comp.get_ismultiplier())
                                               and not (comp.getshifter())):

                    # only those components that can get convoluted and are
                    # not shifters or multipliers
                    comp.calculate()
                    if comp.gethasmultiplier():  # if it has a multiplier
                        multi = comp.getmultiplierptr()

                        if multi is not None:
                            dummyspec = comp.copy()
                            # multiply the result with the multiplier component
                            dummyspec *= multi
                            # add the multiply result to what we have already
                            self += dummyspec

                    else:
                        self += comp  # just add the component

        # then apply the convolution if needed
        if use_ll:
            for comp in self.components:
                if isinstance(comp, Mscatter):
                    comp.data = np.ndarray.copy(
                        self.data)  # copy in component the model data so far
                    comp.calculate()  # convolve

                    self.data = np.ndarray.copy(
                        comp.data)  # and copy it back in the model

        # then add components which can not be convoluted and are not
        # convolutors and are not shifters and are not multipliers
        for comp in self.components:
            if not (comp.getcanconvolute()) and not (comp.get_ismultiplier()) \
                    and not isinstance(comp, Mscatter) \
                    and not (comp.getshifter()):
                comp.calculate()
                self += comp

    def getconvolutor(self):
        if not self.hasconvolutor():
            return None

        for comp in self.components:
            if isinstance(comp, Mscatter):
                return comp

    def resetsigmas(self):
        """
        Reset the error bars on all parameters.

        Returns
        -------
        None.

        """
        for comp in self.components:
            for p in comp.parameters:
                p.setsigma(0.0)

    def plot(self, spectrum=None, externalplt=None, **kwargs):
        """
        Plot the Model

        Parameters
        ----------
        spectrum: Spectrum
            A spectrum can be plotted simultaneously
        externalplt : matplotlib reference
              A reference to an external matplotlib reference, if None we use
              our own matplotlib and create a new figure.
        """
        self.calculate()
        tempplt = plt
        if isinstance(externalplt, plt.Figure):
            tempplt = externalplt
        else:
            # create our own figure
            plt.figure()
            plt.title('Model')
        # show components if visible
        for comp in self.components:
            tempplt.plot(self.energy_axis, comp.data, **kwargs)
        tempplt.plot(self.energy_axis, self.data,
                     **kwargs)  # and the total spectrum
        tempplt.xlabel(r'Energy Loss [eV]')
        if spectrum is not None:
            tempplt.plot(spectrum.energy_axis, spectrum.data, color='black',
                         label='Experimental data')
        tempplt.ylabel('Counts')
        tempplt.legend()

    def componentshow(self):
        """
        show contents of all component in a separate graph if they are visible
        """
        for comp in self.components:
            comp.show()

    def printcomponents(self):
        """
        Print a list summarising all components.

        Returns
        -------
        None.

        """
        for comp in self.components:
            print('name: ', comp.getname())

    def __iadd__(self, spectrum):
        """
        += operator, only add the spectral data.

        Parameters
        ----------
        spectrum : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        self.data += spectrum.data
        return self

    def __imul__(self, spectrum):
        """
        *= operator, only multiply the spectral data

        Parameters
        ----------
        spectrum : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        self.data *= spectrum.data
        return self

    def seteshift(self, energy):
        # override of Spectrum::seteshift
        # set eshift for the model
        # super().seteshift(energy)
        super().eshift = energy
        # do the same for all components, they also get a setchanged command so
        # they will recalculate completely when calculate is called
        for comp in self.components:
            # comp.seteshift(energy)
            comp.eshift = energy
        # every new component that gets added needs also to get the same eshift
        # this is done in addcomponent}

    def hasconvolutor(self):
        for comp in self.components:
            if comp._isconvolutor:
                return True
        return False

    def hasshifter(self):
        for comp in self.components:
            if comp._isshifter:
                return True

        return False

    def hasmultiplier(self):
        for comp in self.components:
            if comp._ismultiplier:
                return True

        return False

    def getgradient(self, parameter):
        # ask for an analytical gradient if available
        if parameter.gethasgradient():
            # analytical gradients don't work if the model contains convolutor
            # shifters or multipliers
            # we could apply these to the individual analytical gradients but
            # this takes almost the same time
            # as a numerical derivative

            # find the component
            component = self.getcomponentbyparameter(parameter)
            if component is None:
                return None

            if self.hasmultiplier() or self.hasconvolutor() \
                    or self.hasshifter():
                return None

            # TODO get gradient from a parameter,
            #  create it in component for convenience
            return component.getgradient(parameter)
        else:
            return None



import sys

import numpy as np
from pyEELSMODEL.core.spectrum import Spectrum, Spectrumshape
from pyEELSMODEL.core.model import Model

from pyEELSMODEL.components.gaussian import Gaussian
from pyEELSMODEL.fitters.lsqfitter import LSQFitter

sys.path.append("..")  # Adds higher directory to python modules path.


def test_perform_fit():
    specshape = Spectrumshape(1, 100, 1024)
    pol = Gaussian(specshape, 10000, 400, 50)
    mod = Model(specshape, components=[pol])
    mod.calculate()
    s = Spectrum(specshape, data=np.random.normal(mod.data, 1))

    # values = [1,110,1]
    values = [10000, 400, 50]

    for param, value in zip(pol.parameters, values):
        param.setvalue(value)

    fit = LSQFitter(s, mod)
    fit.perform_fit()

    for value, coeff in zip(values, fit.coeff):
        rel_er = np.abs((coeff - value) / value)
        assert rel_er < 1e-2


def test_use_bounds():
    specshape = Spectrumshape(1, 100, 1024)
    pol = Gaussian(specshape, 10000, 400, 50)
    mod = Model(specshape, components=[pol])
    mod.calculate()
    s = Spectrum(specshape, data=np.random.normal(mod.data, 1))

    fit = LSQFitter(s, mod, use_bounds=True)
    assert not fit.use_bounds

    fit = LSQFitter(s, mod, method='trf', use_bounds=True)
    assert fit.use_bounds


def main():
    test_perform_fit()
    test_use_bounds()


if __name__ == "__main__":
    main()



from pyEELSMODEL.misc import hydrogen_gdos as hdos
from pyEELSMODEL.components.CLedge.coreloss_edge import CoreLossEdge


class HydrogenicCoreLossEdge(CoreLossEdge):
    """
      Calculates the core loss cross-section using the hydrogenic wave
      function. The cross-sections can be calculated for the K and L edges.

      """

    def __init__(self, specshape, A, E0, alpha, beta, element, edge, eshift=0,
                 q_steps=100):
        """

        Parameters
        ----------
        specshape : Spectrumshape
        The spectrum shape used to model
        A : float
        Amplitude of the edge

        E0: float [V]
        The acceleration voltage of the microscope

        alpha: float [rad]
        The convergence angle of the incoming probe

        beta: float [rad]
        The collection angle

        element: string
        The name of the element from which to calculate the edge model.

        edge: string
        The type of edge. (K, L or M)

        eshift: float [eV]
        The shift of the onset energy with respect to the literature value.
        (default: 0)

        q_steps: int
        The number of q points taken into account for the integration over
        the momentum space. The larger the number of q_steps the more
        accurate the calculation. (default: 100)

        Returns
        -------
        """

        super().__init__(specshape, A, E0, alpha, beta, element, edge,
                         eshift=eshift, q_steps=q_steps)

    def set_onset_energy(self):
        if self.edge == 'L':
            self.onset_energy = hdos.get_IE3()[self.Z - 13] + self.eshift

            if self.Z == 6:
                self.onset_energy = 5 + self.eshift

        elif self.edge == 'K':
            self.onset_energy = hdos.get_EK()[self.Z - 1] + self.eshift

    def set_edge(self, edge):
        """
        Checks if it is possible to calculate the given edge. If so, then
        this edge is set as an attribute.

        Parameters
        ----------
        edge : string
              The edge which needs to calculated.

        """
        if not isinstance(edge, str):
            raise TypeError('Edge should be a string: K or L')

        if edge == 'K' or edge == 'L':
            if edge == 'L':
                if (self.Z < 13):
                    # added the oxgyen L edge out of curiosity
                    raise ValueError('No L edge for this element')

                elif self.Z > 36:
                    raise ValueError('Hydrogenic L edge not supported for '
                                     'this element (Z>36)')
                else:
                    self.edge = edge
            else:
                self.edge = edge
        else:
            raise ValueError('Edge should be K or L for hydrogenic method')

    def calculate_cross_section(self):
        """
        Calculates the cross section using the hydrogenic wavefunctions.

        Returns
        -------
        cross_section: 1d numpy array [barn]
            The calculated cross section using the hydrogenic wavefunctions.

        """
        csc = 1e28 * hdos.dsigma_dE_hydrogenic(self.energy_axis, self.Z,
                                               self.onset_energy,
                                               self.parameters[1].getvalue(),
                                               self.parameters[2].getvalue(),
                                               self.parameters[3].getvalue(),
                                               shell=self.edge,
                                               q_steps=self.q_steps)

        return csc



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
import numpy as np
from pyEELSMODEL.core.component import Component
from pyEELSMODEL.core.parameter import Parameter
from pyEELSMODEL.core.spectrum import Spectrum
import logging

logger = logging.getLogger(__name__)


class FixedPattern(Component):
    """
    Fixed pattern as component. This component can be used when an experimental
    edge can get extracted from experimental data.

    Parameters
    ----------
    specshape : Spectrumshape
        The spectrum shape used to model

    spectrum: Spectrum
        The spectrum from which to extract the fixed pattern.

    A : float
        The amplitude of fixed pattern compared to the data of
        the spectrum. (default: 1)

    scale: float (> 0)
        A scale value for changing the energy axis. This can be used
        when another dispersion is used between the fixed pattern and
        the experimental data.
        (default: 1)

    shift: float [eV]
        Shifting the data from the spectrum with the given value.
        (default: 0)

    Returns
    -------

    """

    def __init__(self, specshape, spectrum, A=1, scale=1, shift=0, name=' '):
        super().__init__(specshape)
        p1 = Parameter('A', A)
        p1.setlinear(True)
        p1.setboundaries(-np.inf, np.inf)
        p1.sethasgradient(True)
        self._addparameter(p1)

        p2 = Parameter('scale', scale, changeallowed=False)
        self._addparameter(p2)

        p3 = Parameter('shift', shift, changeallowed=False)
        self._addparameter(p3)

        self.set_fixeddata(spectrum)

        self._spectrum = Spectrum(self.get_spectrumshape(),
                                  data=self.fixeddata)

        self.setdisplayname(name)
        self.setdescription("Fixed pattern of " + name)
        self._setname(name)

    def set_fixeddata(self, spectrum):
        """
        It can happen that the fixed pattern does not have the same energy axis
        as the spectrum to which we want to fit. Hence a interpolation is
        performed to match the fixedpattern data with the energy axis of the
        model.

        Parameters
        ----------
        spectrum: Spectrum
            The spectrum from where the data is interpolated on the
            other energy axis.

        Returns
        -------
        int_spec.data: numpy 1d array
            The raw data of the interpolated spectrum on the energy axis
            given by the spectrum shape of the fixed pattern.

        """
        int_spec = spectrum.interp_to_other_energy_axis(self)
        self.fixeddata = int_spec.data[:]
        # return int_spec.data[:]

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        p3 = self.parameters[2]

        if p1.ischanged() or p2.ischanged() or p3.ischanged():
            A = p1.getvalue()
            scale = p2.getvalue()
            shift = p3.getvalue()
            self.data = self.modified_fixedpattern(self.fixeddata, A, scale,
                                                   shift)
        self.setunchanged()  # put parameters to unchanged

    def modified_fixedpattern(self, fixedpattern, A, scale, shift):
        """
        Recalculates the fixed pattern. The amplitude,

        Parameters
        ----------
        fixedpattern: numpy array (1D)
            The unmodified fixed pattern data.
        A: float
            The amplitude of the fixed pattern.
        scale: float
            The scale to change the energy axis
        shift: float
            The shift of the energy axis

        Returns
        -------
        res: numpy array (1D)
            The recalculated data from the fixed pattern.


        """
        if (scale == 1) and (shift == 0):
            res = A * fixedpattern
        else:
            res = A * self._spectrum.rescale_spectrum(scale, shift).data
        return res

    def getgradient(self, parameter):
        pA = self.parameters[0]
        if parameter == pA:
            self.gradient[0] = self.fixeddata
            return self.gradient[0]
        else:
            return None



# -*- coding: utf-8 -*-
"""
Created on Thu Dec  9 20:46:09 2021

@author: joverbee
"""
import numpy as np
from scipy import interpolate
import logging
from pyEELSMODEL.operators.aligns.align import Align
from tqdm import tqdm

logger = logging.getLogger(__name__)


class AlignCrossCorrelation(Align):
    """
    Aligns the dataset using the maximum from the cross correlation of the
    given spectra with the reference. At this point the reference is taken as
    the spectrum which is set in the multispectrum. This alignment method will
    only work if an edge or zero loss is the same over the
    entire scan.


    """
    def __init__(self, multispectrum, other_spectra=[], signal_range=None,
                 cropping=False, interp=1, subpixel=False, is_zlp=False):
        """
        Parameters
        ----------
        multispectrum: Multispectrum
            The spectrum from which the background should be removed
        other_spectra: List of Multispectra
            The other multispectra which can be aligned with the same
            parameters
        signal_range: tuple
            Indicates the region on which the cross correlation should be
            performed.
            (default: None)
        cropping: bool
            Indicates if the edges of the signal are cropped out when aligning
             the zero loss peak.
            (default: False)
        interp: int >1
            The interpolation used for subpixel accuracy (default: 1)
        is_zlp: boolean
            Inidicates if the aligned spectrum is a zero loss peak. If so,
            the offset is modified at the end to put maximum value at zero
            energy
        """

        super().__init__(multispectrum, other_spectra, cropping,
                         signal_range=signal_range, zero_index=None)
        self._reference = np.copy(self.multispectrum.data)
        # normalization of reference
        self._reference = self._reference - self._reference.mean()
        self.interp = interp
        # also other interpolation kinds can be used.
        self.interpolation_kind = 'linear'
        self.subpixel = subpixel
        self.method = 'Cross correlation'
        self.is_zlp = is_zlp

    @property
    def reference(self):
        """
        Reference spectrum used in the cross correlation
        """
        return self._reference

    @reference.setter
    def reference(self, ref):
        """
        Sets the reference data which is also normalized to use in the cross
        correlation
        """
        ind0 = self.multispectrum.get_energy_index(self.signal_range[0])
        ind1 = self.multispectrum.get_energy_index(self.signal_range[1])
        if self.subpixel:
            ref_ = ref[ind0:ind1] - ref[ind0:ind1].mean()
            self._reference = self.interpolate_int(ref_, self.interp)
        else:
            self._reference = ref[ind0:ind1] - ref[ind0:ind1].mean()

    @property
    def correlationmap(self):
        return self._correlationmap

    @correlationmap.setter
    def correlationmap(self, correlationmap):
        self._correlationmap = correlationmap

    @property
    def interp(self):
        return self._interp

    @interp.setter
    def interp(self, index):
        if index < 1 or not isinstance(index, int):
            raise ValueError(r'The given interpolation step should be an '
                             r'integer larger than 0')
        self._interp = index

    @property
    def subpixel(self):
        return self._subpixel

    @subpixel.setter
    def subpixel(self, b):
        self._subpixel = b
        ref = np.copy(self._reference)
        self.reference = ref

    def interpolate_int(self, data, index):
        """
        Function which interpolates the data with the number of index points.
        So if index is two, the interpolated data will contain twice as many
        points.

        Parameters
        -------
        data:

        index:

        Returns
        -------
        None.
        """
        old_axes = np.linspace(0, 1, data.size)
        new_axes = np.linspace(0, 1, data.size * index + 1 - index)
        f = interpolate.interp1d(old_axes, data, kind=self.interpolation_kind)
        return f(new_axes)

    def determine_shift(self):
        """
        Sets the index_shift and shift attribute by finding the maximum value
        of the cross correlation between each spectrum and the reference
        spectrum. Reference spectrum is fixed and should be chosen beforehand.

        Returns
        -------
        None.

        """
        ind0 = self.multispectrum.get_energy_index(self.signal_range[0])
        ind1 = self.multispectrum.get_energy_index(self.signal_range[1])
        shape = (self.multispectrum.xsize, self.multispectrum.ysize)
        shift = np.zeros(shape)
        correlation_map = np.zeros(shape)

        for index in tqdm(np.ndindex(shape)):
            islice = np.s_[index]
            avg = self.multispectrum.multidata[islice][ind0:ind1].mean()
            signal = self.multispectrum.multidata[islice][ind0:ind1] - avg
            if self.subpixel:
                signal = self.interpolate_int(signal, self.interp)
            # croscor = np.correlate(self.reference[ind0:ind1], signal, 'full')

            croscor = np.correlate(self.reference, signal, 'full')

            shift[islice] = np.argmax(croscor) - self.reference.size + 1
            correlation_map[islice] = np.max(croscor)

        self.index_shift = shift.astype('int')
        if self.subpixel:
            self.shift = -1 * self.index_shift *\
                         self.multispectrum.dispersion/self.interp
        else:
            self.shift = -1*self.index_shift*self.multispectrum.dispersion
        self.correlationmap = correlation_map

    def perform_alignment(self):
        """
        Performs the alignment procedure by first calculating the shift and
        then applying it to the spectra. The roll method is chosen to perform
        the alignment.

        Returns
        -------
        None.

        """
        self.determine_shift()
        self.align()

        if self.is_zlp:
            zlp_pos = self.aligned.energy_axis[
                np.argmax(self.aligned.mean().data)]

            self.aligned.offset -= zlp_pos

            for spec in self.aligned_others:
                spec.offset -= zlp_pos

        # if self.subpixel:
        #     self.align()
        # else:
        #     self.fast_align()



# -*- coding: utf-8 -*-
"""
Created on Thu Dec  9 20:46:09 2021

@author: joverbee
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backend_bases import MouseButton
import logging

from pyEELSMODEL.core.operator import Operator
from pyEELSMODEL.core.multispectrum import MultiSpectrum
from pyEELSMODEL.components.CLedge.coreloss_edge import CoreLossEdge

logger = logging.getLogger(__name__)


class CalibrateSpectrum(Operator):
    """
    The CalibrateSpectrum class provides a workflow to change the energy axis
    which is sometimes wrong. It does this by performing a linear fit between
    the measured onset energies (defined by user) and values obtained from
    literature.
    """

    def __init__(self, spectrum, element_list, edge_list, other_spectra=[]):
        """
        Creates CalibrateSpectrum object.

        Parameters
        ----------
        spectrum: Spectrum
            The spectrum used for the calibration
        element_list: List of str
            List which contains the different elemtens seen on the EEL spectrum
        edge_list: List of str
            List which contains the different edges seen on the EEL spectrum
        other_spectra: List of Spectra of MultiSpectra
            List on which the same type of correction should be applied

        """
        self.spectrum = spectrum.copy()
        self.element_list = element_list
        self.edge_list = edge_list

        spectra = []
        for other in other_spectra:
            spectra.append(other.copy())

        self.other_spectra = spectra

        self.set_onset_edge_energies()

    def set_onset_edge_energies(self):
        """
        Calculates the list of tabulated onset energies of the edges.
        """
        self.onset_edge_energies = []

        sh = self.spectrum.get_spectrumshape()
        for elem, edge in zip(self.element_list, self.edge_list):
            core = CoreLossEdge(sh, 1, 1, 1, 1, elem, edge)
            self.onset_edge_energies.append(core.onset_energy)

    def show_real_energies(self):
        """
        Show the real energies from the tabulated data on the spectrum
        """

        cte = 1.1
        fig, ax = plt.subplots()
        if isinstance(self.spectrum, MultiSpectrum):
            ndata = self.spectrum.sum().data
        else:
            ndata = self.spectrum.data

        ax.plot(self.spectrum.energy_axis, ndata)
        for index, en in enumerate(self.onset_edge_energies):
            lab = self.element_list[index]+' '+self.edge_list[index]+' edge'
            ax.plot([en, en], [ndata.min(), cte*ndata.max()], label=lab)

        ax.set_ylim([ndata.min(), cte*ndata.max()])
        ax.set_xlabel(r'Energy axis')
        ax.legend()

    def determine_edge_positions(self):
        """
        Use user input by using right mouse click to select the edge energy.
        After the click is performed, it is shown which edge you had selected
        and you can verify if this is correct.

        :return:
        """

        def onclick(event):
            if event.button is MouseButton.LEFT:
                pass
            else:
                global ix, iy
                ix, iy = event.xdata, event.ydata
                ylim = ax.get_ylim()
                ax.plot([event.xdata, event.xdata], [ylim[0], ylim[1]],
                        color='black')
                ax.set_ylim(ylim)
                fig.canvas.draw()
                print('Energy of ' + self.element_list[len(coords)] + ' ' +
                      self.edge_list[len(coords)] + ' is:' + str(ix))

                coords.append(ix)
                if len(coords) == len(self.onset_edge_energies):
                    fig.canvas.mpl_disconnect(cid)

                return coords

        print('right click the edge')
        cte = 1.1
        fig, ax = plt.subplots()
        if isinstance(self.spectrum, MultiSpectrum):
            ndata = self.spectrum.sum().data
        else:
            ndata = self.spectrum.data
        ax.plot(self.spectrum.energy_axis, ndata)
        ax.set_ylim([ndata.min(), cte*ndata.max()])
        ax.set_xlabel(r'Energy axis')
        global coords
        coords = []
        self.coords = coords
        cid = fig.canvas.mpl_connect('button_press_event', onclick)

    def calibrate_energy_axis(self):
        """
        Calibrates the energy axis using the input energies.

        """
        if len(self.coords) != len(self.onset_edge_energies):
            print('Run the determine_edge_positions function '
                  'again to define the edges')

        self.coeff = np.polyfit(np.array(self.coords),
                                np.array(self.onset_edge_energies), 1)
        n_en_axis = self.coeff[0] * self.spectrum.energy_axis + self.coeff[1]
        disp = n_en_axis[1]-n_en_axis[0]
        offset = n_en_axis[0]
        delta_offset = self.spectrum.offset - offset

        self.spectrum.dispersion = disp
        self.spectrum.offset = offset

        for spectrum in self.other_spectra:
            spectrum.dispersion = disp
            spectrum.offset -= delta_offset

    def show_calibration_fit(self):
        plt.figure()
        plt.plot(self.coords, self.onset_edge_energies,
                 marker='o', label='Data points')
        plt.plot(self.spectrum.energy_axis,
                 self.coeff[0] * self.spectrum.energy_axis + self.coeff[1],
                 label='Fit')
        plt.legend()



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
from pyEELSMODEL.core.component import Component
from pyEELSMODEL.core.parameter import Parameter
import numpy as np
from pyEELSMODEL.misc import physical_constants as pc
import math


class Plasmon(Component):
    """
    A physical model of ZL and plasmon peaks This is not very adquate to use
    for model based fitting. The description of the plasmon calculation can be
    found in the Egerton book, Edition 3 (p.139, formula 3.43c)

    """

    def __init__(self, specshape, A, Ep, Ew, n, tlambda, beta, E0):
        """
        Parameters
        ----------
        specshape: Spectrumshape
            Spectrumshape on which the plasmon should be calculated
        A: float
            Integral of the zero-loss peak
        Ep: float [eV]
            Position plasmon energy
        Ew: float [eV]
            The width of the plasmon peak
        n: uint
            The number of plasmons
        tlambda: float
            The inelastic mean free path
        beta: float [rad]
            The collection angle
        E0: float [V]
            The acceleration voltage

        """

        super().__init__(specshape)

        p0 = Parameter('A', A)
        p0.setlinear(True)
        p0.setboundaries(0, np.inf)
        p0.sethasgradient(False)
        self._addparameter(p0)  # integral zer

        p1 = Parameter('Ep', Ep, changeallowed=False)  # eV
        p1.sethasgradient(False)
        self._addparameter(p1)

        p2 = Parameter('Ew', Ew, changeallowed=False)  # eV
        p2.sethasgradient(False)
        self._addparameter(p2)

        p3 = Parameter('Eb', 1,
                       changeallowed=False)
        p3.sethasgradient(False)
        self._addparameter(p3)  # height zero loss peak

        p4 = Parameter('n', n, changeallowed=False)
        p4.sethasgradient(False)
        self._addparameter(p4)

        p5 = Parameter('tlambda', tlambda, changeallowed=False)
        p5.sethasgradient(False)
        self._addparameter(p5)

        p6 = Parameter('beta', beta, changeallowed=False)
        p6.sethasgradient(False)
        self._addparameter(p6)

        p7 = Parameter('E0', E0, changeallowed=False)
        p7.sethasgradient(False)
        self._addparameter(p7)  # height zero loss peak

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        self.data = self.plasmon_function()

    def plasmon_function(self):
        for i in range(self.parameters[4].getvalue()):
            fraction = self.poisson_fraction(i + 1,
                                             self.parameters[5].getvalue())
            plasmon = fraction * self.calculate_plasmon(
                self.parameters[1].getvalue() * (i + 1),
                self.parameters[2].getvalue(), self.parameters[3].getvalue(),
                self.parameters[6].getvalue(), self.parameters[7].getvalue())
            if i == 0:
                tot_plasmon = np.copy(plasmon)
            else:
                tot_plasmon += plasmon

        return (np.exp(self.parameters[5].getvalue()) - 1) * self.parameters[
            0].getvalue() * tot_plasmon / tot_plasmon.sum()

    def calculate_plasmon(self, Ep, Ew, Eb, beta, E0):
        """
        The drude model of the plasmon from which the enerrgy differential
        cross section is calculated. The description of the formula can be
        found in the Egerton book (p.139, formula 3.43c)

        Parameters
        ----------
        Ep: float [eV]
            Position plasmon energy
        Ew: float [eV]
            The width of the plasmon peak
        Eb: float [eV]
            The binding energy of plasmon
        beta: float [rad]
            The collection angle
        E0: float [V]
        """
        E = self.energy_axis
        eps = 1 - Ep ** 2 / (E ** 2 - Eb ** 2 + E * Ew * 1j)
        inv_eps = np.imag(-1 / eps)

        theta_E = E / (pc.joule_to_eV(
            pc.gamma(E0) * pc.m0() * pc.speed_electron(E0) ** 2))

        prefactor = (np.pi * (1e9 * pc.a0()) * pc.joule_to_eV(
            pc.m0() * pc.speed_electron(
                E0) ** 2)) ** -1  # nm, borh radius in nm
        anglog = np.zeros(theta_E.size)
        anglog[E > 0] = np.log(1. + beta ** 2 / theta_E[E > 0] ** 2)
        volint = prefactor * inv_eps * anglog
        volint[E <= 0] = 0  # else it will give back nan, which gives trouble
        return volint

    def poisson_fraction(self, n, tlambda):
        """

        :param n: integer value of the plasmon
        :param tlambda: thickness over mean free path (lambda) ratio
        :return: fraction
        """
        nfac = math.factorial(n)
        fraction = np.exp(-tlambda) * tlambda ** n / nfac
        return fraction



"""
Testing the functionality of the multiple zero loss alignements.
"""

import numpy as np
import pytest

from pyEELSMODEL.core.multispectrum import MultiSpectrum, MultiSpectrumshape
from pyEELSMODEL.core.spectrum import Spectrumshape
from pyEELSMODEL.components.gaussian import Gaussian
from pyEELSMODEL.components.lorentzian import Lorentzian
from pyEELSMODEL.operators.aligns.alignzeroloss import AlignZeroLoss
from pyEELSMODEL.operators.aligns.fastalignzeroloss import FastAlignZeroLoss
from pyEELSMODEL.operators.aligns.align import Align
from pyEELSMODEL.operators.aligns.aligncrosscorrelation import \
    AlignCrossCorrelation


# Make a dataset where the centre changes.
# Every column has a incremental increase and between every row there is
# a random offset

def make_artificial_dataset(start, end, model_type="Gaussian",
                            return_shift=False):
    """
    Make artificial datasets used to test the performance of the zero loss
    alignment
    """

    ncol = 100  # number of columns
    nrow = 12  # number of rows
    width = 10  # widht of the centre misalignement
    fwhm = 3  # fwhm of gaussian or lorentzian
    amp = 100  # amplitude
    specshape = Spectrumshape(0.5, -200, 1024)

    centre_array = np.linspace(start * width, end * width, ncol)
    offs_array = 10 * np.random.rand(nrow)

    sim_data = np.zeros((nrow, ncol, specshape.size))
    sim1_data = np.zeros(
        sim_data.shape)  # extra spectra which will be aligned using zl
    sim2_data = np.zeros(
        sim_data.shape)  # extra spectra which will be aligned using zl
    real_shift = np.zeros((nrow, ncol))
    nind = 200  # the number of indices which are high
    for i in range(nrow):
        for j in range(ncol):
            if model_type == 'Gaussian':
                zl = Gaussian(specshape, amp, centre_array[j] + offs_array[i],
                              fwhm)
                zl1 = Gaussian(specshape, amp,
                               50 + centre_array[j] + offs_array[i], fwhm)
                zl2 = Gaussian(specshape, amp,
                               100 + centre_array[j] + offs_array[i], fwhm)

            else:
                zl = Lorentzian(specshape, amp,
                                centre_array[j] + offs_array[i], fwhm)
                zl1 = Lorentzian(specshape, amp,
                                 50 + centre_array[j] + offs_array[i], fwhm)
                zl2 = Lorentzian(specshape, amp,
                                 100 + centre_array[j] + offs_array[i], fwhm)

            zl.calculate()
            zl1.calculate()
            zl2.calculate()

            sim_data[i, j] = zl.data

            sim_data[i, j, :nind] = 1
            sim_data[i, j, -nind:] = 5
            sim1_data[i, j] = zl1.data
            sim1_data[i, j, :nind] = 1
            sim1_data[i, j, -nind:] = 5

            sim2_data[i, j] = zl2.data
            sim2_data[i, j, :nind] = 1
            sim2_data[i, j, -nind:] = 5
            real_shift[i, j] = centre_array[j] + offs_array[i]

    mshape = MultiSpectrumshape(specshape.dispersion, specshape.offset,
                                specshape.size, nrow, ncol)
    s = MultiSpectrum(mshape, data=sim_data)
    s1 = MultiSpectrum(mshape, data=sim1_data)
    s2 = MultiSpectrum(mshape, data=sim2_data)
    if return_shift:
        return s, s1, s2, real_shift
    else:
        return s, s1, s2


def test_init_other():
    s, _, _ = make_artificial_dataset(-2, 2)

    # wrong xsize
    mshape = MultiSpectrumshape(1, 10, 1024, s.xsize + 20, s.ysize)
    s1 = MultiSpectrum(mshape)
    with pytest.raises(ValueError):
        Align(s, other_spectra=[s1], cropping=True)

    # wrong ysize
    mshape = MultiSpectrumshape(1, 10, 1024, s.xsize, s.ysize + 20)
    s1 = MultiSpectrum(mshape)
    with pytest.raises(ValueError):
        Align(s, other_spectra=[s1], cropping=True)


def test_fast_align():
    shf_l = [[-2, 2], [-4, -2], [2, 4]]
    for i in range(len(shf_l)):
        s, s1, s2 = make_artificial_dataset(shf_l[i][0], shf_l[i][1])
        align = FastAlignZeroLoss(s)
        align.perform_alignment()
        # every maximum value is at zero energy
        assert np.all(align.aligned.energy_axis[
                          np.argmax(align.aligned.multidata, axis=2)].astype(
            'int') == 0)


def test_fast_align_1():
    """
    Align the fast zero loss using the align function which does the
    interpolation instead of roll
    """
    shf_l = [[-2, 2], [-4, -2], [2, 4]]
    for i in range(len(shf_l)):
        s, s1, s2 = make_artificial_dataset(shf_l[i][0], shf_l[i][1])
        align = FastAlignZeroLoss(s, other_spectra=[s1, s2])
        align.determine_fast_shift()
        align.align()
        # every maximum value is at zero energy
        alg = align.aligned
        alg0 = align.aligned_others[0]
        alg1 = align.aligned_others[1]

        Em = alg.energy_axis[np.argmax(alg.multidata, axis=2)]
        Em0 = alg0.energy_axis[np.argmax(alg0.multidata, axis=2)]
        Em1 = alg1.energy_axis[np.argmax(alg1.multidata, axis=2)]

        assert np.all(np.abs(Em)) <= 1
        assert np.all(np.abs(Em0-50)) <= 1
        assert np.all(np.abs(Em1-100)) <= 1


def test_fast_align_crop():
    shf_l = [[-2, 2], [-4, -2], [2, 4]]
    for i in range(len(shf_l)):
        s, s1, s2 = make_artificial_dataset(shf_l[i][0], shf_l[i][1])
        align = FastAlignZeroLoss(s, other_spectra=[s1, s2], cropping=True)
        align.perform_alignment()

        assert np.min(align.aligned.multidata[:, :, 0]) == 1
        assert np.max(align.aligned.multidata[:, :, 0]) == 1
        assert np.min(align.aligned.multidata[:, :, -1]) == 5
        assert np.max(align.aligned.multidata[:, :, -1]) == 5

        assert np.min(align.aligned_others[0].multidata[:, :, 0]) == 1
        assert np.max(align.aligned_others[0].multidata[:, :, 0]) == 1
        assert np.min(align.aligned_others[0].multidata[:, :, -1]) == 5
        assert np.max(align.aligned_others[0].multidata[:, :, -1]) == 5

        assert np.min(align.aligned_others[1].multidata[:, :, 0]) == 1
        assert np.max(align.aligned_others[1].multidata[:, :, 0]) == 1
        assert np.min(align.aligned_others[1].multidata[:, :, -1]) == 5
        assert np.max(align.aligned_others[1].multidata[:, :, -1]) == 5


def test_fit_align():
    shf_l = [[-2, 2], [-4, -2], [2, 4]]
    for i in range(len(shf_l)):
        s, s1, s2 = make_artificial_dataset(shf_l[i][0], shf_l[i][1])
        align = AlignZeroLoss(s)
        align.perform_alignment()

        # every maximum value is almost at zero
        assert np.all(np.abs(
            s.energy_axis[np.argmax(align.aligned.multidata, axis=2)]) < 2)


def test_fit_align_lorentz():
    shf_l = [[-2, 2], [-4, -2], [2, 4]]
    for i in range(len(shf_l)):
        s, s1, s2 = make_artificial_dataset(shf_l[i][0], shf_l[i][1],
                                            model_type='Lorentzian')
        align = AlignZeroLoss(s, model_type='Lorentzian')
        align.perform_alignment()

        # every maximum value is almost at zero
        assert np.all(np.abs(
            s.energy_axis[np.argmax(align.aligned.multidata, axis=2)]) < 2)


def fit_align_other():
    s, s1, s2 = make_artificial_dataset(-2, 2)
    align = AlignZeroLoss(s, other_spectra=[s1, s2])
    align.perform_alignment()

    # every maximum value is almost at zero
    assert np.all(
        np.abs(s.energy_axis[np.argmax(align.aligned.multidata, axis=2)]) < 2)
    assert np.all(np.abs((s1.energy_axis[
        np.argmax(align.aligned_others[0].multidata, axis=2)]) - 50) < 2)
    assert np.all(np.abs((s2.energy_axis[
        np.argmax(align.aligned_others[1].multidata, axis=2)]) - 100) < 2)


def fit_align_crop():
    shf_l = [[-2, 2], [-4, -2], [2, 4]]
    for i in range(len(shf_l)):
        s, s1, s2 = make_artificial_dataset(shf_l[i][0], shf_l[i][1])
        align = AlignZeroLoss(s, other_spectra=[s1, s2], cropping=True)
        align.perform_alignment()

        assert np.min(align.aligned.multidata[:, :, 0]) == 1
        assert np.max(align.aligned.multidata[:, :, 0]) == 1
        assert np.min(align.aligned.multidata[:, :, -1]) == 5
        assert np.max(align.aligned.multidata[:, :, -1]) == 5

        assert np.min(align.aligned_others[0].multidata[:, :, 0]) == 1
        assert np.max(align.aligned_others[0].multidata[:, :, 0]) == 1
        assert np.min(align.aligned_others[0].multidata[:, :, -1]) == 5
        assert np.max(align.aligned_others[0].multidata[:, :, -1]) == 5

        assert np.min(align.aligned_others[1].multidata[:, :, 0]) == 1
        assert np.max(align.aligned_others[1].multidata[:, :, 0]) == 1
        assert np.min(align.aligned_others[1].multidata[:, :, -1]) == 5
        assert np.max(align.aligned_others[1].multidata[:, :, -1]) == 5


def test_cross_correlation():
    shf_l = [[-2, 2], [-4, -2], [2, 4]]
    for i in range(len(shf_l)):
        s, s1, s2, real_shift = make_artificial_dataset(shf_l[i][0],
                                                        shf_l[i][1],
                                                        return_shift=True)
        align = AlignCrossCorrelation(s, other_spectra=[s1, s2])
        align.perform_alignment()

        diff = np.abs(align.shift - (real_shift - real_shift[0, 0]))
        assert np.all(diff < 1)


def test_cross_correlation_signal_range():
    shf_l = [[-2, 2], [-4, -2], [2, 4]]
    for i in range(len(shf_l)):
        s, s1, s2, real_shift = make_artificial_dataset(shf_l[i][0],
                                                        shf_l[i][1],
                                                        return_shift=True)
        align = AlignCrossCorrelation(s, other_spectra=[s1, s2],
                                      signal_range=(-50, 50))
        align.perform_alignment()

        diff = np.abs(align.shift - (real_shift - real_shift[0, 0]))
        print(diff)
        assert np.all(diff < 1)


def main():
    test_init_other()
    test_fast_align()
    test_fast_align_1()
    test_fast_align_crop()
    test_fit_align()
    test_fit_align_lorentz()
    fit_align_other()
    fit_align_crop()
    test_cross_correlation()
    test_cross_correlation_signal_range()


if __name__ == "__main__":
    main()



import h5py
import numpy as np


def load_h5py(filename):
    f = h5py.File(filename, 'r')
    d = f['Energy_axis']
    dispersion = d.attrs['Dispersion']
    size = d.attrs['Size']
    offset = d.attrs['Offset']
    data = f['Data'][:]
    f.close()
    return data, dispersion, offset, size


def load_elements_and_edges(filename):
    f = h5py.File(filename, 'r')
    d = f['Energy_axis']
    edges = d.attrs['edges']
    elements = d.attrs['elements']
    f.close()
    return elements, edges


def load_hspy(filename):
    f = h5py.File(filename, 'r')
    d = f['Experiments']
    params = []
    haadf = []
    df = []
    detector_type = []

    for key in d.keys():
        if 'EELS Spectrum' in key:
            data = d[key]['data'][:]
            axis_num = data.ndim - 1

            if data.ndim == 2:
                data = data[:, np.newaxis, :]
            data = data.astype(float)

            ax_key = 'axis-' + str(axis_num)
            offset = d[key][ax_key].attrs['offset']
            dispersion = d[key][ax_key].attrs['scale']
            size = data.shape[-1]
            stra = 'Acquisition_instrument'
            acq_time = d[key]['metadata'][stra]['TEM'].attrs['exposure']
            params.append([data, dispersion, offset, size, acq_time])
        elif 'HAADF' in key:
            data = d[key]['data'][:]
            haadf.append(data)
            detector_type.append(key)
        elif 'DF4' in key:
            data = d[key]['data'][:]
            df.append(data)
            detector_type.append(key)
    f.close()

    if len(haadf) > 0:
        return params, haadf, detector_type
    elif len(df) > 0:
        return params, df, detector_type
    else:
        return params, df, detector_type



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
import numpy as np
from scipy.optimize import minimize

from pyEELSMODEL.core.fitter import Fitter
import logging
logger = logging.getLogger(__name__)


class MinimizeFitter(Fitter):
    """
    Fitter class which uses the minimization algorithms of scipy for
    finding the optimal parameters. Mainly for the maximum likelihood
    estimator but others can also be implemented.

    """

    def __init__(self, spectrum, model, method='nelder-mead', estimator='ML'):
        super().__init__(spectrum, model)
        self.estimator = estimator
        self.method = method

    @property
    def estimator(self):
        return self._estimator

    @estimator.setter
    def estimator(self, estimator):
        if estimator == 'ML':
            self._estimator = self.MLE_function
            self.jacobian = self.jacobian_ML

        # if estimator is 'ML' and not self.model.hasconvolutor():
        #     self._estimator = self.MLE_function
        #     self.jacobian = self.jacobian_ML

        # elif estimator is 'ML' and self.model.hasconvolutor():
        #     self._estimator = self.MLE_function
        #     self.jacobian = '2-point'

        elif estimator == 'LSQ':
            self._estimator = self.lsq_function
            self.jacobian = '2-point'
        else:
            logger.warning(r'Estimator is not known, maximum'
                           r' likelihood is chosen')
            self._estimator = self.MLE_function
            self.jacobian = self.jacobian_ML

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, method):
        self._method = method

    @property
    def jacobian(self):
        """
        :return:
        """
        return self._jacobian

    @jacobian.setter
    def jacobian(self, jacobian):
        """
        Only when the gradients are used, the jacobian will be used.
        :param jacobian:
        :return:
        """
        if self.usegradients:
            self._jacobian = jacobian
        else:
            self._jacobian = '2-point'

    def MLE_function(self, x0, m, s):
        """
        The maximum likelihood estimator
        :param x0:
        :param m:
        :param s:
        :return:
        """
        for index, param in enumerate(m.getfreeparameters()):
            param.setvalue(x0[index])
        m.calculate()

        if any(m.data < 0):
            return np.inf

        # boolean1 = (s.data > 1.) & (m.data > 1.)
        # when the experimental dat is smaller than zero, then the
        # stirling approx will give wrong result
        boolean1 = m.data > 0
        boolean2 = np.invert(s.exclude)
        boolean = np.logical_and(boolean1, boolean2)

        # paper where factor two is used
        # (https://doi.org/10.1016/S0168-9002(00)00756-7)
        ML = 2*np.sum(s.data[boolean] * np.log(m.data[boolean])
                      - m.data[boolean])
        return -ML

    def jacobian_ML(self, x0, m, s):
        # todo can still have errors
        for index, param in enumerate(m.getfreeparameters()):
            param.setvalue(x0[index])
        m.calculate()

        if any(m.data < 0):
            return np.inf

        boolean1 = m.data > 0
        boolean2 = np.invert(s.exclude)
        boolean = np.logical_and(boolean1, boolean2)

        jac = []
        for index, param in enumerate(m.getfreeparameters()):
            comp = m.getcomponentbyparameter(param)
            gradient = comp.getgradient(param)
            j = 2 * np.sum((1 - s.data[boolean]/m.data[boolean])
                           * (gradient[boolean]))
            jac.append(j)

        return jac

    def jacobian_LSQ(self, x0, m, s):
        # todo needs to be implemented

        return 1

    def lsq_function(self, x0, m, s):
        """
        Least squares minimization, preferably use LSQFitter
        """
        for index, param in enumerate(m.getfreeparameters()):
            param.setvalue(x0[index])
        m.calculate()
        inv_ = np.invert(s.exclude)
        res = np.sum(np.abs((s.data[inv_] - m.data[inv_])))
        return res

    def perform_fit(self):
        """
        Does the fit on a single spectrum where the starting parameters are
        given by x0.

        """
        x0 = self.get_start_parameters()
        bounds = self.get_bounds()
        argo = (self.model, self.spectrum)

        resML = minimize(self.estimator, x0, args=argo, method=self.method,
                         tol=1e-12, bounds=bounds, jac=self.jacobian)

        self.coeff = resML.x



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
from pyEELSMODEL.core.spectrum import Spectrum
from pyEELSMODEL.core.parameter import Parameter
import logging

logger = logging.getLogger(__name__)


class Component(Spectrum):
    """
    A Component is a Spectrum that can calculate itself from a number of
    parameters. Base class from which all components can be derived with
    specific functionality

    """

    def __init__(self, spectrumshape, params=None):
        """
        Instantiate a Component object

        Parameters
        ----------
        spectrumshape : Spectrumshape
            Describing the size of the spectrum this component will generate,
            see info on Spectrum.
        params : list of Parameter, optional
            A list of parameters controlling this component
            This is typically passed when loading a previously made component
            from a file. If a new component is constructed, typically the
            component will create and take care of its own parameters. The
            default is None.

        Returns
        -------
        None.

        """
        super().__init__(spectrumshape)
        self.parameters = []
        self.gradient = []
        self._setparameters(
            params)  # use preloaded parameters if given in params
        self.spectrumshape = spectrumshape

        
        self._isshifter = False # indicates a special type of component that can shift the energies
        self._multiplier = None
        self._ismultiplier = False
        self._hasmultiplier = False
        self._isconvolutor = False
        self.displayname = ""
        self.description = ""
        self._name = ""
        self.visible = True
        self.suppress = False

        self._setshifter(False)
        self._setcanconvolute(True)
        self._set_ismultiplier(False)
        self.setvisible(False)
        self.setdisplayname("")
        self.setdescription("unitialized component")
        self._setname("unitialized component")

    def _addparameter(self, parameter):
        """
        Add parameter to a component and indicate if the component
        provides an analytical gradient for this parameter.
        This is an internal function meant to be used only in the init of a
        specific component

        Parameters
        ----------
        parameter : Parameter
            A parameter to add to the component.

        Returns
        -------
        None.

        """
        self.parameters.append(parameter)
        self.gradient.append(Spectrum(self.spectrumshape))

    def release(self):
        # do all you need to do to tell that we are no longer there
        # release all parameters
        # (eg in case they might be coupled to other parameters)

        for p in self.parameters:
            p.release()
    
    def setsuppress(self,b):
        """
        Sets the suppress state of the component. Allows to switch of a component in the sense that calling
        calculate will return a zero spectrum. Use with caution as forgetting to disable this function
        may lead to hard to interpret behaviour

        Parameters
        ----------
        b : bool
            suppress state.

        Returns
        -------
        None.

        """
        self.suppress=b
        self.setchanged()
        #if you have parameters that are coupled, than also that component needs to know that something changed
        for p in self.parameters:
            if p.iscoupled():
                cp=p.getcontroller()
                cp.setchanged()


    def setvisible(self, b):
        """
        Sets the visible state of the component. Can indicate whether in a
        model spectrum showing individual components, this component needs to
        be visible or not. default True.

        Parameters
        ----------
        b : bool
            visible state.

        Returns
        -------
        None.

        """
        self.visible = b
        self.show()  # show it if it was set to true

    def _setname(self, name):
        """
        Sets the name of the component. This name needs to be unique for each
        component as it will be used to load components from file when loading
        a model.

        Parameters
        ----------
        name : string
            The name of the component.

        Returns
        -------
        None.

        """
        self.name = name
        # also give that name to the spectrum that is at the base of the
        # component
        super().setname(name)

    def setdisplayname(self, name):
        """
        Sets the displayed name of the component. This name can differ from the
        actual name of the component and allows the user to change the name to
        a more meaningful name without affecting the official name that should
        not change.

        Parameters
        ----------
        name : string
            Display name for the component.

        Returns
        -------
        None.

        """
        self.displayname = name
        super().setname(name)

    def setdescription(self, description):
        """
        Set the description string of the component. This is typically the
        business of the component itself, but can be changed by the user. The
        description can be used e.g. inthe UI to provide help explaining the
        function of the component

        Parameters
        ----------
        description : string
            A description of what the component does.

        Returns
        -------
        None.

        """
        self.description = description

    def getname(self):
        """
        Get the official and fixed name of a component. This name is a unique
        identifier of the type of component. This name should not be changed
        by the user.

        Returns
        -------
        string
            Unque identifier name of the component.

        """
        return self.name

    def getdisplayname(self):
        """
        Get the friendly name of a component. This name can be changed by the
        user and will typically be used for designating the component in eg. a
        user interface.

        Returns
        -------
        string
            A name describing this component.

        """
        return self.displayname

    def getdescription(self):
        """
        Get a string describing the functionality of the component.
        Can be used in e.g. help or tooltips in the UI.

        Returns
        -------
        string
            String describing the functionality.

        """
        return self.description

    def _setparameters(self, params):
        """
        Set the list of Parameter relevant to this component. Note that
        typically a component has a well defined set of parameters and the
        user should not need to tamper with these. This function is used when
        loading a component from file and sending it a previously defined
        list of parameters.

        Parameters
        ----------
        params : list of Parameter
            list of parameters which needs to be added to the component.

        Returns
        -------
        None.

        """
        if params is None:
            return
        else:
            for p in params:
                self._addparameter(p)  # this also reserves space for gradients

    def print_parameter_values(self):
        """
        Print all parameters of a component as name : value

        Returns
        -------
        None.

        """
        for p in self.parameters:
            print(p.getname(), ' : ', p.getvalue())

    def setunchanged(self):
        """
        Set all parameters of a Component as unchanged. This is typically done
        internally after a component.calculate() but can also be called
        externally to avoid that a component will be recalculated if fiddling
        with some parameters externally. Note however that if you call this
        function, the next time you call recalculate the component will falsely
        assume that it doesnt need to do anything as the parameters are marked
        as unchanged!

        Returns
        -------
        None.

        """
        for p in self.parameters:
            p.setunchanged()

    def setchanged(self):
        """
        Mark all parameters as changed. This notifies the component that on the
        next component.calculate() call, the full component needs to be
        recalculated. Note that it even marks unchangeable Parameters as
        changed even though their value doesnt.

        Returns
        -------
        None.

        """
        for p in self.parameters:
            p.setchanged()

    def ischanged(self):
        changed = False
        for p in self.parameters:
            changed = changed or p.ischanged()
        return changed

    def has_parameter(self, p):
        """
        Search if Parameter p is one of the parameters in the parameter list of
         this component.

        Parameters
        ----------
        p : Parameter
            Reference to input parameter to find in the parameter list of this
            component.

        Raises
        ------
        TypeError
            In case p is not a Parameter object.

        Returns
        -------
        bool
            True if parameter p indeed is one of the parameters of this
            component, False otherwise.

        """
        if isinstance(p, Parameter):
            return p in self.parameters
        else:
            raise TypeError(r'Input should be Parameter object')

    def _setcanconvolute(self, b):
        """
        Set the canconvolute state of the component. Default=True This
        indicates whether this Component should be modified by Convoluter type
        components when calculating the result of a Model. A typical use case
        is for an EELS background component that would not improve when
        convolved with a low loss spectrum while it leads to nasty edge
        artefacts. The solution is to not convolve this component at all which
        can be obtained by setting this setcanconvolute state to False

        Parameters
        ----------
        b : bool
            canconvolute state.

        Returns
        -------
        None.

        """
        self.canconvolute = b

    def getcanconvolute(self):
        """
        Gets wether this component can be convoluted

        Parameters
        ----------
        None.

        Returns
        -------
        None.

        """
        return self.canconvolute

    def _setshifter(self, b):
        """
        Sets whether this component is a Shifter type component (b=True).
        Default is False. A shifter Component can alter the energy values of
        the model to mimick the effect of experimental energy drift or eg.
        nonlinearity of the dispersion. All 'normal' components should not
        alter this. The model will treat shifter components in a special way
        when calculating

        Parameters
        ----------
        b : bool
            isshifter state.

        Returns
        -------
        None.

        """
        self.isshifter = b

    def getshifter(self):
        """
        Get whether this is a shifter component

        Returns
        -------
        bool
            True if shifter type component.

        """
        return self.isshifter

    def calculate(self):
        """
        This calculates the spectral data of the component. Each derived class
        from Component will need to implement its own functionality here.

        Returns
        -------
        None.

        """
        logger.warning(
            'if you see this you have called calculate on a '
            'component that didnt implement calculate')
        # raise NotImplementedError() #this forces a derived class to
        # implement it

    def seteshift(self, eshift):
        """
        Sets an energy shift for this component which shifts its energy values
        with respect to the energy axis given when instantiating the component.
        This can be used to mimick experimental energy drift. In order to make
        sure the component gets a full recalculation at the next calculate()
        call, all parameters will be indicated as changed.

        Parameters
        ----------
        eshift : float
            energy shift.

        Returns
        -------
        None.

        """
        super().seteshift(
            eshift)  # shift the energy of the underlying spectrum
        # make sure to recalculate as the energy values have changed
        self.setchanged()

    def getparameter(self, index):
        """
        Return the parameter with index.

        Parameters
        ----------
        index : int
            Index to indicate which parameter to return
        Raises
        ------
        ValueError
            When index out of bounds.
        TypeError
            When index is not int

        Returns
        -------
        Parameter
            Reference to the requested Parameter in the params list.

        """
        if self.indexOK(index):
            return self.parameters[index]
        else:
            raise IndexError('bad index')
            return None

    def indexOK(self, index):
        """
        Check if index is within range of params list

        Parameters
        ----------
        index : int
            index.

        Returns
        -------
        bool
            True: index is good,, False otherwise.

        """
        return 0 <= index < len(self.parameters)

    def getmultiplier(self):
        """
        Return a reference to a Multiplier Component if one is attached to this
        component Otherwise returns None. Multiplier components can be used
        e.g. to mimick effect of gain variation of a camera.

        Raises
        ------
        AttributeError
            DESCRIPTION.

        Returns
        -------
        multiplier : Component
            Reference to another component which is indicated to be a
            multiplier of this component.

        """
        if not self._hasmultiplier:
            return None
        try:
            self.multiplier.get_ismultiplier()
        except Exception:
            raise AttributeError('Invalid reference to a multiplier component')
            self._hasmultiplier = False
            self.multiplier = None
        return self.multiplier

    def gethasmultiplier(self):
        return self.hasmultiplier

    def sethasmultiplier(self, b):
        self.hasmultiplier = b

    def setmultiplier(self, component):
        """
        Set a multiplier component to this component. The component needs to
        be of the multiplier type. The multiplier component will take the
        output of this component and multiply it with its content

        Parameters
        ----------
        component : Component
            The component which needs to be set as multiplier.
        Raises
        ------
        AttributeError
            DESCRIPTION.

        Returns
        -------
        None.

        """
        dummy = False
        try:
            dummy = component.get_ismultiplier()
        except Exception:
            self.hasmultiplier = False
            self.multiplier = None
            raise AttributeError('invalid component')
        if dummy:
            self.multiplier = component
            self.hasmultiplier = True
        else:
            self.multiplier = None
            self.hasmultiplier = False

    def _set_ismultiplier(self, b):
        self.ismultiplier = b
        # a multiplier can not have a multiplier itself
        self.hasmultiplier = False
        self.multiplier = None

    def get_ismultiplier(self):
        return self._ismultiplier

    def releasemultiplier(self):
        # tell the component that we no longer have a multiplier
        self.hasmultiplier = False
        self.multiplier = None

    def getfreelinparameters(self):
        freelinparameters = []
        for p in self.parameters:
            if p.ischangeable():
                if p.islinear():
                    freelinparameters.append(p)
        return freelinparameters

    def _pullparameter(self):
        """
        Pull last parameter from parameter list and releases that parameter.
        This effectively reduces the number of parameters of a component
        NOT sure why we would need this???and definitely not a user function

        Returns
        -------
        None.

        """
        # remove last parameter from the parameter list
        try:
            p = self.parameters.pop()
            p.release()
        except Exception:
            return

    def show(self):
        """
        Plot component contents if the component has isvisible==True

        Returns
        -------
        None.

        """
        if self.visible:
            self.plot()

    def getgradient(self, parameter):
        """
        Get a reference to the gradient of this component with respect to
        parameter, if an analytical gradient is available. This function needs
        to be overoaded for every specific component

        Parameters
        ----------
        parameter : Parameter
            Reference to an existing parameter in the component.

        Returns
        -------
        A reference to a Spectrum holding the gradient if the parameter was
        indeed a parameter of this component and has an analytical gradient
        defined gradient. None in all other cases.

        """
        return None

    def save(self, fh):
        # save the details of this component to file in such a way
        # that you can bring it to the same state later

        # save the file location of the ll spectrum to open it later
        # when reloading the model

        # TODO: probably XML with elementtree would be most logica
        # component can override this if they need to store more details

        return

    def load(self, fh):
        # load this component from saved detaile

        # TODO: probably XML with elementtree would be most logica
        # component can override this if they need to store more details
        return



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
import sys
import copy
import logging

logger = logging.getLogger(__name__)


class Parameter():
    """
    Parameter class to store a parameter to be used in e.g. a component that
    typically has a list of parameters. Parameters hold a single numerical
    value, can have lower and upper boundaries, can be locked can be coupled to
    other parameters, can have the 'linear' property, indicating that in a
    component this parameter is linear as opposed to a nonlinear parameter that
    affects the component in a nonlinear way. Parameters can also be monitored
    by a Monitor class and they can be watched by a Watcher class
    """

    def __init__(self, name, val, changeallowed=True):
        """
        Instantiates a Parameter
        Parameters
        ----------
        name : string
            A meaningful name for the parameter.
        val : float or int
            The initial value of the parameter.
        changeallowed : bool, optional
            Wether the parameter can be changed after instantiating.
            The default is True.

        Returns
        -------
        None.

        """
        self.setdefaults()
        self.setname(name)
        self.setvalue(val)
        self.setchangeable(changeallowed)
        logger.debug('Parameter init')

    def setdefaults(self):
        """
        Sets default values, typically only called in __init__

        Returns
        -------
        None.

        """
        self.value = 0
        self.changed = True
        self.displayed = False
        self.changeable = True
        self.couple_fraction = 1.0
        self.bound = False
        self.monitored = False
        self.watched = False
        self.linear = False
        self.watcher = None
        self.coupled = False
        self.coupled_parameter = None
        self.monitor = 0
        self.setname("")
        self.sigma = 0.0
        self.setboundaries(-sys.float_info.max, sys.float_info.max)
        self.hasgradient = False
        self.valid = True

    def release(self):
        """
        Call this function whenever you no longer need this parameter
        it warns other coupled Parameter, Monitor and Watcher instances
        that might be linked to it that they should no longer count on this
        parameter.

        Returns
        -------
        None.

        """
        self.valid = False

        self.releasecoupling()
        # self.releasemonitor()
        # self.releasewatcher()

    def sethasgradient(self, b):
        self.hasgradient = b

    def gethasgradient(self):
        return self.hasgradient

    def copy(self):
        """
        Returns a copy of the object
        :return:
        """
        return copy.deepcopy(self)

    def setvalue(self, val):
        """
        Sets the value of the parameter. But only if it is not a coupled slave,
        if this is changeable and if the new value is within the bounds if the
        parameter is bound.

        Parameters
        ----------
        val : float
            New value of the parameter.

        Returns
        -------
        True if value was set.

        """
        # only change when changeable and within boundaries,
        # otherwise do nothing...
        if self.iscoupled():
            logger.info('This parameters cannot be changed since it is '
                        'coupled')
            return False  # this parameter cant'be changed

        # if self.coupled and self.ismaster():
        #     #mark the slave as changed
        #     self.coupled_parameter.setchanged() #important!!!

        okbd = self.boundaryOK(val, self.getlowerbound(), self.getupperbound())
        if self.changeable and okbd:
            self.value = val
            self.changed = True
        if not okbd:
            # try:
            #     v = float(val)
            # except:
            #     logger.warning('invalid float (prob infinity')
            #     return False
            logger.info(
                'Hitting the boundary on parameter %s, ub=%e, lb=%e,'
                ' requested val=%e. either increase the boundary or'
                ' try to find out why the fit is diverging',
                self.getname(), self.getupperbound(),
                self.getlowerbound(), val)
            return False
        return True

    def getvalue(self):
        """
        Gets the current value of the Parameter. If the parameter is a coupled
        slave, the value of the master times a coupling constant will be
        returned

        Returns
        -------
        float
            Value of parameter or of the master parameter times a coupling
             constant if this is a coupled slave.

        """
        if self.iscoupled():
            # get the value of the coupled parameter if the master parameter
            # still exists
            coupledval = self.coupled_parameter.getvalue()
            return coupledval * self.couple_fraction
        # in normal case, just return the value
        else:
            return self.value

    def interactivevalue(self, description):  # interactive setup of the value
        """
        Not implemented.
        """

        # not clear what to do here
        return

    def setdisplayed(self, b):
        """
        Set the displayed state of the parameter. This can be used from a UI
        to indicate if the UI has to be informed if e.g. something changes.

        Parameters
        ----------
        b : bool
            True if displayed on a UI, False otherwise.

        Returns
        -------
        None.

        """
        if b:
            self.displayed = True
        else:
            self.displayed = False

    def isdisplayed(self):
        """
        Return whether the parameter is marked as 'displayed'. This can be set
         with setdisplayed(True/False)

        Returns
        -------
        bool
            True is displayed state is set, False otherwise.

        """
        return self.displayed

    @property
    def sigma(self):
        """
        Get the error bar on this parameter.

        Returns
        -------
        sig : float
            value of the error bar for this parameter.
        """
        return self._sigma

    @sigma.setter
    def sigma(self, sig):
        """
        Sets the error bar on this parameter. This is merely a variable that
        stores any number you put in it, however it is meant as a storage for
        a calculated precision estimate on the parameter.

        Parameters
        ----------
        sig : float
            value of the error bar for this parameter.

        Returns
        -------
        None.
        """
        self._sigma = sig

    def setunchanged(self):
        """
        Sets the parameter to the unchanged state. This can be used by the
        model to signal that the model has already been calculated with these
        parameters and we don't need to repeat the calculation unless someone
        changes the parameter value.
        """
        self.changed = False

    def setchanged(self):
        """
        Force the parameter to a changed state.
        This is useful to e.g. set a coupled slave parameter to the changed
        state even though that parameter is typically unchangeable as it should
        follow the master. This is important as a component needs to know that
        this parameter has changed in order to force a full recalculation when
        recalculate is called. This can also be used to force a recalculate of
        a Component as changed parameters will avoid that the Component takes
        a shortcut because it thinks it already calculated with this parameter
        value.

        Returns
        -------
        None.

        """
        # this is a warning received by a slave of coupling
        # that parameter has changed
        # if self.iscoupled() and self.isslave():

        self.changed = True  # even if a slave is unchangeable...

    def setboundaries(self, lb, ub, force=False):
        """
        Sets numerical boundaries to the parameter but only if the current
        value fits within those boundaries. The boundaries are inclusive >=
        and <=. Swapping lower for upper boundary is allowed and will
        automatically be corrected. If force is True, the value of the
        parameter will be changed to make the boundaries work.

        Parameters
        ----------
        lb : float
            lower boundary.
        ub : float
            upper boundary.
        force: bool
            If force is true, then the boundaries are set even if the value
            of the parameter falls outside this boundary. The new value of the
            parameter will be set in the center of the boundaries.
            (default: True)


        Returns
        -------
        bool
            True if bound, False otherwise.

        """
        # check if value fits in range, if not, don't set the boundaries
        # return false if bounds are not set
        self.bound = True
        if (self.boundaryOK(self.value, lb, ub)):
            self.lowerbound = lb
            self.upperbound = ub
            return True
        if (self.boundaryOK(self.value, ub, lb)):
            self.lowerbound = lb
            self.upperbound = ub
            return True
        if force:
            self.lowerbound = lb
            self.upperbound = ub
            val = (ub + lb) / 2
            self.setvalue(val)

        logger.info(r'Boundaries are not set')
        self.bound = False
        return self.bound

    def setupperbound(self, ub):
        """
        Sets the upper numerical bound on the parameter. No checking is
        performed on whether the current value would violate this bound

        Parameters
        ----------
        ub : float
            upper boundary.

        Returns
        -------
        None.

        """
        self.upperbound = ub

    def setlowerbound(self, lb):
        """
        Sets the lower numerical bound on the parameter. No checking is
        performed on whether the current value would violate this bound

        Parameters
        ----------
        ub : float
            lower boundary.

        Returns
        -------
        None.

        """
        self.lowerbound = lb

    def boundaryOK(self, x, lb, ub):
        """
        Check if a proposed value x is within the bounds given by lower bound
        (lb) and upper bound (ub).

        Parameters
        ----------
        x : float
            value to be tested.
        lb : float
            lower bound.
        ub : float
            upper bound.

        Returns
        -------
        bool
            returns True if x is within bounds OR if the parameter is not
            bound.

        """
        return not (self.bound) or ((x >= lb) and (x <= ub))

    def getupperbound(self):
        """
        Returns the current upper bound.

        Returns
        -------
        float
            upper bound value.

        """
        return self.upperbound

    def getlowerbound(self):
        """
        Returns the current lower bound.

        Returns
        -------
        float
            lower bound value.

        """
        return self.lowerbound

    def setchangeable(self, b):
        """
        Set the changeable state of the parameter. If not changeable, the
        parameter value can not be updated. This can be used to indicate to a
        Model which are free parameters and which parameters are fixed.

        Parameters
        ----------
        b : bool
            True if parameter can be changed, False if it can't be changed.

        Returns
        -------
        None.

        """
        if b:
            # only allow to set to changeable if the parameter is not coupled
            if self.iscoupled():
                logger.warning('Attempt to set a coupled parameter to '
                               'changeable, ignoring this request')
            else:
                self.changeable = True
        else:
            self.changeable = False

    def setlinear(self, b):
        """
        Sets a boolean variable to indicate whether a parameter is linear or
        not. This can be used e.g. by the Fitter to know if a pure linear Model
        (all free parameters are linear) is present which allows to use a much
        faster fitting algorithm.

        Parameters
        ----------
        b : bool
            True for linear parameter, False otherwise.

        Returns
        -------
        None.

        """
        self.linear = b

    def islinear(self):
        """
        Returns whether this parameter is indicated as linear. This is used
        e.g. by a model to check if a linear fit can be used (all free
        parameters are linear) instead of a nonlinear.

        Returns
        -------
        bool
            True if parameter is indicated as linear, False otherwise.

        """
        return self.linear

    def ischanged(self):
        """
        Returns wether a parameter is indicated as changed (typically since
        last time eg. a Component used it for calculation).

        Returns
        -------
        bool
            True if parameter has changed, False otherwise.

        """
        if self.coupled and self.coupled_parameter is not None:
            return self.coupled_parameter.changed
        else:
            return self.changed

    def ischangeable(self):
        """
        Returns wether a parameter is changeable with setvalue().

        Returns
        -------
        bool
            True if parameter can be changed, False otherwise.

        """
        if self.iscoupled():
            return False  # this parameter cant'be changed
        else:
            return self.changeable

    def isbound(self):
        """
        Returns whether a parameter has active bounds.

        Returns
        -------
        bool
            True if parameter is bound, False otherwise.

        """
        return self.bound

    def iscoupled(self):
        """
        Returns whether a parameter is coupled to another parameter. Checks are
        performed to see if the coupled parameter is still valid.

        Returns
        -------
        bool
            True if coupled, False otherwise.

        """
        if self.coupled and self.coupled_parameter is not None:
            try:
                if not self.coupled_parameter.valid:
                    raise ValueError('invalid coupling reference in Parameter')
            except Exception:
                logger.warning('coupled parameter doesnt respond, deleting the'
                               'coupling')
                self.releasecoupling()
                return False
            return True
        else:
            return False

    def getcontroller(self):
        """
        Get a reference to the controller parameter if this is a coupled
        worker.

        Returns
        -------
        Parameter
            reference to a master Parameter instance if we are a coupled
            worker.
        None
            if we are not a coupled worker.

        """
        if self.iscoupled():
            return self.coupled_parameter
        else:
            return None

    def setname(self, name):
        """
        Sets the name of the parameter. name can be any string, preferably a
        meaningful description
        of the parameter

        Parameters
        ----------
        name : string
            name of the parameter.

        Returns
        -------
        None.

        """
        self.name = name

    def getname(self):
        """
        Get name string including "coupled to:" in case this parameter is
        coupled.

        Returns
        -------
        string
            the name of the parameter including 'coupled to...' in case the
            parameter is coupled.

        """
        if self.iscoupled():
            totalname = ' '.join((self.name, ':coupled to ',
                                  self.coupled_parameter.getname()))
            return totalname
        else:
            return self.name

    def printall(self):
        """
        Print some details on the parameter, mostly for debugging and testing.

        Returns
        -------
        None.

        """
        print(self.getname(), ' val=', self.getvalue(), ' changeable=',
              self.ischangeable(), ' lb=', self.lowerbound,
              ' ub=', self.upperbound)

    def getpurename(self):
        """
        Get the 'pure' name of the parameter. This is the exact name as given
        with setname() or during init() and doesnt include any info on eventual
        coupling to another parameter.

        Returns
        -------
        string
            The name of the parameter.

        """
        return self.name

    def couple(self, parameter, fraction=1.0):
        """
        Couples this parameter to another parameter with a certain coupling
        constant. This will make us into slaves of the master parameter that is
        passed. We no longer will be changeable and a call to getvalue() will
        result in the value of the controller parameter.

        Parameters
        ----------
        parameter : Parameter
            reference a another parameter to couple to.
        fraction : float, optional
            coupling constant. The default is 1.0.

        Returns
        -------
        None.

        """

        logger.info('coupling parameter %s with: %s', self.getname(),
                    parameter.getname())
        self.coupled = True
        self.coupled_parameter = parameter  # pointer to the master
        self.couple_fraction = fraction
        # test the master parameter to see if it is valid
        try:
            if not parameter.valid:
                raise ValueError('parameter no longer valid')
        except Exception:
            # something went wrong
            self.releasecoupling()
            logger.warning('failed to access the parameter')
            return

        # check if the proposed parameter is not yourself
        # otherwise recursion! and crash
        if (self == parameter):
            # something went wrong
            logger.warning('recursive coupling detected\n')
            self.releasecoupling()
            return
        self.setchangeable(
            False)
        # we no longer need to be changeable by the user since from now on
        # the master determines our value
        logger.info('done coupling')

    def getcouplingfactor(self):
        """
        Return the current value of the coupling factor, irrespective of
        whether the parameter is actually coupled or not.

        Returns
        -------
        float
            value of the coupling factor.

        """
        return self.couple_fraction

    def __sub__(self, other):
        """
        Subtract parameters.
        If other is an int or float, the number is subtracted from the value of
        the self parameter. The result has no predefined name and default
        (widest) boundaries.

        Parameters
        ----------
        other : Parameter or int or float
            The parameter or int or float to subtract.

        Raises
        ------
        TypeError
            In case other is not a Parameter or float or int.

        Returns
        -------
        s : Parameter
            Returns a reference to a -new- Parameter that holds the subtraction
            result.

        """
        if type(other) is Parameter:
            s = Parameter('', self.getvalue() - other.getvalue(),
                          changeallowed=True)
            return s
        elif type(other) is int or type(other) is float:
            s = Parameter('', self.getvalue() - other, changeallowed=True)
            return s

        else:
            raise TypeError('other should either be a Parameter, int or '
                            'float.')

    def __add__(self, other):
        """
        Add parameters. If other is an int or float, the number is added to the
        value of the self parameter. The result has no predefined name and
        default (widest) boundaries.

        Parameters
        ----------
        other : Parameter or int or float
            The parameter or int or float to add.

        Raises
        ------
        TypeError
            In case other is not a Parameter or float or int.

        Returns
        -------
        s : Parameter
            Returns a reference to a -new- Parameter that holds the addition
            result.

        """
        if type(other) is Parameter:
            s = Parameter('', self.getvalue() + other.getvalue(),
                          changeallowed=True)
            return s
        elif type(other) is int or type(other) is float:
            s = Parameter('', self.getvalue() + other, changeallowed=True)
            return s

        else:
            raise TypeError('other should either be a Parameter, int or '
                            'float.')

    def __mul__(self, other):
        """
        Multiply parameters. If other is an int or float, the number multiplied
        with the value of the self parameter. The result has no predefined
        name and default (widest) boundaries.

        Parameters
        ----------
        other : Parameter or int or float
            The parameter or int or float to multiply.

        Raises
        ------
        TypeError
            In case other is not a Parameter or float or int.

        Returns
        -------
        s : Parameter
            Returns a reference to a -new- Parameter that holds the
            multiplication result.

        """

        if type(other) is Parameter:
            s = Parameter('', self.getvalue() * other.getvalue(),
                          changeallowed=True)
            return s
        elif type(other) is int or type(other) is float:
            s = Parameter('', self.getvalue() * other, changeallowed=True)
            return s

        else:
            raise TypeError('other should either be a Parameter, int or '
                            'float.')

    def __truediv__(self, other):
        return self.__div__(other)

    def __div__(self, other):
        """
        Divide parameter by another parameter or float or int. The resulting
        Parameter has no predefined name and default (widest) boundaries.

        Parameters
        ----------
        other : Parameter or int or float
            The parameter or int or float to divide.

        Raises
        ------
        TypeError
            In case other is not a Parameter or float or int.

        Returns
        -------
        s : Parameter
            Returns a reference to a -new- Parameter that holds the division
            result.

        """
        if type(other) is Parameter:
            s = Parameter('', self.getvalue() / other.getvalue(),
                          changeallowed=True)
            return s
        elif type(other) is int or type(other) is float:
            s = Parameter('', self.getvalue() / other, changeallowed=True)
            return s
        else:
            raise TypeError('other should either be a Parameter, int or '
                            'float.')

    def releasecoupling(self):
        """
        Stops any coupling of this parameter with another

        Returns
        -------
        None.

        """
        self.coupled = False
        self.coupled_parameter = None
        self.setchangeable(True)  # not sure if this is needed, in principle
